import math
from typing import Mapping

from .arch import ATD

from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len, get_pixelshuffle_params, pixelshuffle_scale


class ATDArch(Architecture[ATD]):
    def __init__(self):
        super().__init__(
            uid='ATD',
            detect=KeyCondition.has_all(
                'relative_position_index_SA',
                'conv_first.weight',
                'conv_first.bias',
                'layers.0.residual_group.td',
                'layers.0.residual_group.layers.0.sigma',
                'layers.0.residual_group.layers.0.norm1.weight',
                'layers.0.residual_group.layers.0.norm1.bias',
                'layers.0.residual_group.layers.0.norm2.weight',
                'layers.0.residual_group.layers.0.norm2.bias',
                'layers.0.residual_group.layers.0.norm3.weight',
                'layers.0.residual_group.layers.0.norm3.bias',
                'layers.0.residual_group.layers.0.wqkv.weight',
                'layers.0.residual_group.layers.0.attn_win.relative_position_bias_table',
                'layers.0.residual_group.layers.0.attn_win.proj.weight',
                'layers.0.residual_group.layers.0.attn_win.proj.bias',
                'layers.0.residual_group.layers.0.attn_atd.scale',
                'layers.0.residual_group.layers.0.attn_atd.wq.weight',
                'layers.0.residual_group.layers.0.attn_atd.wk.weight',
                'layers.0.residual_group.layers.0.attn_atd.wv.weight',
                'layers.0.residual_group.layers.0.attn_aca.logit_scale',
                'layers.0.residual_group.layers.0.attn_aca.proj.weight',
                'layers.0.residual_group.layers.0.convffn.fc1.weight',
                'layers.0.residual_group.layers.0.convffn.fc1.bias',
                'layers.0.residual_group.layers.0.convffn.dwconv.depthwise_conv.0.weight',
                'layers.0.residual_group.layers.0.convffn.dwconv.depthwise_conv.0.bias',
                'layers.0.residual_group.layers.0.convffn.fc2.weight',
                'layers.0.residual_group.layers.0.convffn.fc2.bias',
                'norm.weight',
                'norm.bias',
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        img_size = 64  # cannot be deduced from state dict
        patch_size = 1  # cannot be deduced from state dict
        img_range = 1.0  # cannot be deduced from state dict

        in_chans = state_dict['conv_first.weight'].shape[1]
        embed_dim = state_dict['conv_first.weight'].shape[0]

        window_size = math.isqrt(state_dict['relative_position_index_SA'].shape[0])

        num_layers = get_seq_len(state_dict, 'layers')
        depths = [6] * num_layers
        num_heads = [6] * num_layers
        for i in range(num_layers):
            depths[i] = get_seq_len(state_dict, f'layers.{i}.residual_group.layers')
            num_heads[i] = state_dict[f'layers.{i}.residual_group.layers.0.attn_win.relative_position_bias_table'].shape[1]

        num_tokens = state_dict['layers.0.residual_group.layers.0.attn_atd.scale'].shape[0]
        reducted_dim = state_dict['layers.0.residual_group.layers.0.attn_atd.wq.weight'].shape[0]
        convffn_kernel_size = state_dict['layers.0.residual_group.layers.0.convffn.dwconv.depthwise_conv.0.weight'].shape[2]
        mlp_ratio = state_dict['layers.0.residual_group.layers.0.convffn.fc1.weight'].shape[0] / embed_dim
        qkv_bias = 'layers.0.residual_group.layers.0.wqkv.bias' in state_dict
        ape = 'absolute_pos_embed' in state_dict
        patch_norm = 'patch_embed.norm.weight' in state_dict

        resi_connection = '1conv' if 'layers.0.conv.weight' in state_dict else '3conv'

        if 'conv_up1.weight' in state_dict:
            upsampler = 'nearest+conv'
            upscale = 4
        elif 'conv_before_upsample.0.weight' in state_dict:
            upsampler = 'pixelshuffle'
            upscale, _ = get_pixelshuffle_params(state_dict, 'upsample')
        elif 'conv_last.weight' in state_dict:
            upsampler = ''
            upscale = 1
        else:
            upsampler = 'pixelshuffledirect'
            upscale = pixelshuffle_scale(state_dict['upsample.0.weight'].shape[0], in_chans)

        norm = 'no_norm' not in state_dict

        is_light = upsampler == 'pixelshuffledirect' and embed_dim == 48
        # use a heuristic for category_size
        category_size = 128 if is_light else 256

        tags = [f'{embed_dim}dim', f'{window_size}w', f'{category_size}cat']
        if is_light:
            tags.insert(0, 'light')

        model = ATD(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
            depths=depths,
            num_heads=num_heads,
            window_size=window_size,
            category_size=category_size,
            num_tokens=num_tokens,
            reducted_dim=reducted_dim,
            convffn_kernel_size=convffn_kernel_size,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            ape=ape,
            patch_norm=patch_norm,
            upscale=upscale,
            img_range=img_range,
            upsampler=upsampler,
            resi_connection=resi_connection,
            norm=norm,
        )

        return self._enhance_model(model=model, in_channels=in_chans, out_channels=in_chans, upscale=upscale, name='ATD')



import math
from typing import Mapping

from .arch import RCAN
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len, get_pixelshuffle_params


class RCANArch(Architecture[RCAN]):
    def __init__(self):
        super().__init__(
            uid='RCAN',
            detect=KeyCondition.has_any(
                KeyCondition.has_all(
                    'head.0.weight',
                    'tail.1.weight',
                    'body.0.body.0.body.0.weight',
                    'body.0.body.0.body.3.conv_du.0.weight',
                ),
                KeyCondition.has_all(
                    'head.1.weight',
                    'tail.1.weight',
                    'body.0.body.0.body.0.weight',
                    'body.0.body.0.body.3.conv_du.0.weight',
                ),
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        n_resgroups = get_seq_len(state_dict, 'body') - 1
        n_resblocks = get_seq_len(state_dict, 'body.0.body') - 1
        head_index = 0
        scale, n_feats = get_pixelshuffle_params(state_dict, 'tail.0')
        unshuffle_mod = get_seq_len(state_dict, 'head') > 1
        n_colors = state_dict['tail.1.weight'].shape[0]
        if unshuffle_mod:
            head_index += 1
            unshuffled_channels = state_dict[f'head.{head_index}.weight'].shape[1]
            downscale_factor = int(math.sqrt(unshuffled_channels / n_colors))
            scale = 4 // downscale_factor
        rgb_range = 255  # undetectable
        kernel_size = state_dict[f'head.{head_index}.weight'].shape[-1]
        norm = 'sub_mean.weight' in state_dict
        reduction = n_feats // state_dict['body.0.body.0.body.3.conv_du.0.weight'].shape[0]
        res_scale = 1  # undetectable
        act_mode = 'relu'  # undetectable
        model = RCAN(
            scale=scale,
            n_resgroups=n_resgroups,
            n_resblocks=n_resblocks,
            n_colors=n_colors,
            rgb_range=rgb_range,
            norm=norm,
            kernel_size=kernel_size,
            n_feats=n_feats,
            reduction=reduction,
            res_scale=res_scale,
            act_mode=act_mode,
            unshuffle_mod=unshuffle_mod,
        )

        return self._enhance_model(model=model, in_channels=n_colors, out_channels=n_colors, upscale=scale, name='RCAN')



import math
from typing import Literal, Sequence

import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor, nn


SampleMods = Literal['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']


class DySample(nn.Module):
    """Adapted from 'Learning to Upsample by Learning to Sample':
    https://arxiv.org/abs/2308.15085
    https://github.com/tiny-smart/dysample
    """

    def __init__(
        self,
        in_channels: int,
        out_ch: int,
        scale: int = 2,
        groups: int = 4,
        end_convolution: bool = True,
    ) -> None:
        super().__init__()

        if in_channels <= groups or in_channels % groups != 0:
            msg = 'Incorrect in_channels and groups values.'
            raise ValueError(msg)

        out_channels = 2 * groups * scale**2
        self.scale = scale
        self.groups = groups
        self.end_convolution = end_convolution
        if end_convolution:
            self.end_conv = nn.Conv2d(in_channels, out_ch, kernel_size=1)

        self.offset = nn.Conv2d(in_channels, out_channels, 1)
        self.scope = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        if self.training:
            nn.init.trunc_normal_(self.offset.weight, std=0.02)
            nn.init.constant_(self.scope.weight, val=0)

        self.register_buffer('init_pos', self._init_pos())

    def _init_pos(self) -> Tensor:
        h = torch.arange((-self.scale + 1) / 2, (self.scale - 1) / 2 + 1) / self.scale
        return torch.stack(torch.meshgrid([h, h], indexing='ij')).transpose(1, 2).repeat(1, self.groups, 1).reshape(1, -1, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        offset = self.offset(x) * self.scope(x).sigmoid() * 0.5 + self.init_pos
        B, _, H, W = offset.shape
        offset = offset.view(B, 2, -1, H, W)
        coords_h = torch.arange(H) + 0.5
        coords_w = torch.arange(W) + 0.5

        coords = (
            torch.stack(torch.meshgrid([coords_w, coords_h], indexing='ij'))
            .transpose(1, 2)
            .unsqueeze(1)
            .unsqueeze(0)
            .type(x.dtype)
            .to(x.device, non_blocking=True)
        )
        normalizer = torch.tensor([W, H], dtype=x.dtype, device=x.device, pin_memory=True).view(1, 2, 1, 1, 1)
        coords = 2 * (coords + offset) / normalizer - 1

        coords = (
            F.pixel_shuffle(coords.reshape(B, -1, H, W), self.scale)
            .view(B, 2, -1, self.scale * H, self.scale * W)
            .permute(0, 2, 3, 4, 1)
            .contiguous()
            .flatten(0, 1)
        )
        output = F.grid_sample(
            x.reshape(B * self.groups, -1, H, W),
            coords,
            mode='bilinear',
            align_corners=False,
            padding_mode='border',
        ).view(B, -1, self.scale * H, self.scale * W)

        if self.end_convolution:
            output = self.end_conv(output)

        return output


class UniUpsample(nn.Sequential):
    def __init__(
        self,
        upsample: SampleMods,
        scale: int = 2,
        in_dim: int = 64,
        out_dim: int = 3,
        mid_dim: int = 64,  # Only pixelshuffle
        group: int = 4,  # Only DySample
    ) -> None:
        m = []

        if scale == 1 or upsample == 'conv':
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'pixelshuffledirect':
            m.extend([nn.Conv2d(in_dim, out_dim * scale**2, 3, 1, 1), nn.PixelShuffle(scale)])
        elif upsample == 'pixelshuffle':
            m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            if (scale & (scale - 1)) == 0:  # scale = 2^n
                for _ in range(int(math.log2(scale))):
                    m.extend([nn.Conv2d(mid_dim, 4 * mid_dim, 3, 1, 1), nn.PixelShuffle(2)])
            elif scale == 3:
                m.extend([nn.Conv2d(mid_dim, 9 * mid_dim, 3, 1, 1), nn.PixelShuffle(3)])
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'nearest+conv':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        (
                            nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                            nn.Upsample(scale_factor=2),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        )
                    )
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            elif scale == 3:
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.Upsample(scale_factor=scale),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'dysample':
            if mid_dim != in_dim:
                m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
                dys_dim = mid_dim
            else:
                dys_dim = in_dim
            m.append(DySample(dys_dim, out_dim, scale, group))
        else:
            raise ValueError(f'An invalid Upsample was selected. Please choose one of {SampleMods}')
        super().__init__(*m)

        self.register_buffer(
            'MetaUpsample',
            torch.tensor(
                [
                    2,  # Block version, if you change something, please number from the end so that you can distinguish between authorized changes and third parties
                    list(SampleMods.__args__).index(upsample),  # UpSample method index
                    scale,
                    in_dim,
                    out_dim,
                    mid_dim,
                    group,
                ],
                dtype=torch.uint8,
            ),
        )


class LayerNorm(nn.Module):
    def __init__(self, dim: int = 64, eps: float = 1e-6) -> None:
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = eps

    def forward(self, x: Tensor) -> Tensor:
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight[:, None, None] * x + self.bias[:, None, None]


class FocusedLinearAttention(nn.Module):
    r"""https://github.com/LeapLabTHU/FLatten-Transformer/blob/master/models/flatten_swin.py
    Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports both of shifted and non-shifted window."""

    def __init__(
        self,
        dim: int = 64,
        window_size: int = 8,
        num_heads: int = 8,
        qkv_bias: bool = True,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        focusing_factor: int = 3,
        kernel_size: int = 5,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.head_dim = head_dim

        self.focusing_factor = focusing_factor
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.softmax = nn.Softmax(dim=-1)

        self.dwc = nn.Conv2d(
            in_channels=head_dim,
            out_channels=head_dim,
            kernel_size=kernel_size,
            groups=head_dim,
            padding=kernel_size // 2,
        )
        self.scale = nn.Parameter(torch.zeros(size=(1, 1, dim)))
        self.positional_encoding = nn.Parameter(torch.zeros(size=(1, window_size**2, dim)))

    def window_partition(self, x: Tensor) -> (Tensor, int, int, int):
        """
        Args:
            x: (B, C, H, W)

        Returns:
            windows: (num_windows*B, window_size, window_size, C)
        """
        B, C, H, W = x.shape
        x = x.view(
            B,
            C,
            H // self.window_size,
            self.window_size,
            W // self.window_size,
            self.window_size,
        )
        windows = x.permute(0, 2, 4, 3, 5, 1).contiguous().view(-1, self.window_size, self.window_size, C)
        return windows, H, W, C

    def window_reverse(self, windows: Tensor, h: int, w: int) -> Tensor:
        B = int(windows.shape[0] / (h * w / self.window_size / self.window_size))
        x = windows.view(
            B,
            h // self.window_size,
            w // self.window_size,
            self.window_size,
            self.window_size,
            -1,
        )
        x = x.permute(0, 5, 1, 3, 2, 4).contiguous().view(B, -1, h, w)
        return x

    def forward(self, x: Tensor) -> Tensor:
        """
        Args:
            x: input features with shape of (num_windows*B, N, C)-
        """

        x, h, w, C = self.window_partition(x)  # nW*B, window_size, window_size, C
        x = x.view(-1, self.window_size * self.window_size, C)  # nW*B, window_size*window_size, C
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, C).permute(2, 0, 1, 3)
        q, k, v = qkv.unbind(0)
        k = k + self.positional_encoding
        q = F.relu(q) + 1e-6
        k = F.relu(k) + 1e-6
        scale = F.softplus(self.scale)
        q = q / scale
        k = k / scale
        q_norm = q.norm(dim=-1, keepdim=True)
        k_norm = k.norm(dim=-1, keepdim=True)
        q = q**self.focusing_factor
        k = k**self.focusing_factor
        q = (q / q.norm(dim=-1, keepdim=True)) * q_norm
        k = (k / k.norm(dim=-1, keepdim=True)) * k_norm

        q = q.reshape(B, N, self.num_heads, -1).permute(0, 2, 1, 3)
        k = k.reshape(B, N, self.num_heads, -1).permute(0, 2, 1, 3)
        v = v.reshape(B, N, self.num_heads, -1).permute(0, 2, 1, 3)

        z = 1 / (q @ k.mean(dim=-2, keepdim=True).transpose(-2, -1) + 1e-6)
        kv = (k.transpose(-2, -1) * (N**-0.5)) @ (v * (N**-0.5))
        x = q @ kv * z

        H = W = self.window_size
        x = x.transpose(1, 2).reshape(B, N, C)
        v = v.reshape(B * self.num_heads, H, W, self.head_dim).permute(0, 3, 1, 2)
        x = x + self.dwc(v).reshape(B, C, N).permute(0, 2, 1)

        x = self.proj(x)
        x = self.proj_drop(x)
        x = self.window_reverse(x, h, w)
        return x


class OmniShift(nn.Module):
    def __init__(self, dim: int = 48) -> None:
        super().__init__()
        # Define the layers for training
        self.conv1x1 = nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=1, groups=dim, bias=True)
        self.conv3x3 = nn.Conv2d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=3,
            padding=1,
            groups=dim,
            bias=True,
        )
        self.conv5x5 = nn.Conv2d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=5,
            padding=2,
            groups=dim,
            bias=True,
        )
        self.alpha1 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)
        self.alpha2 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)
        self.alpha3 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)
        self.alpha4 = nn.Parameter(torch.ones(1, dim, 1, 1), requires_grad=True)

        # Define the layers for testing
        self.conv5x5_reparam = nn.Conv2d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=5,
            padding=2,
            groups=dim,
            bias=True,
        )

    def forward_train(self, x: Tensor) -> Tensor:
        out1x1 = self.conv1x1(x)
        out3x3 = self.conv3x3(x)
        out5x5 = self.conv5x5(x)

        out = self.alpha1 * x + self.alpha2 * out1x1 + self.alpha3 * out3x3 + self.alpha4 * out5x5
        return out

    def reparam_5x5(self) -> None:
        # Combine the parameters of conv1x1, conv3x3, and conv5x5 to form a single 5x5 depth-wise convolution

        padded_weight_1x1 = F.pad(self.conv1x1.weight, (2, 2, 2, 2))
        padded_weight_3x3 = F.pad(self.conv3x3.weight, (1, 1, 1, 1))

        identity_weight = F.pad(torch.ones_like(self.conv1x1.weight), (2, 2, 2, 2))
        combined_weight = (
            self.alpha1.transpose(0, 1) * identity_weight
            + self.alpha2.transpose(0, 1) * padded_weight_1x1
            + self.alpha3.transpose(0, 1) * padded_weight_3x3
            + self.alpha4.transpose(0, 1) * self.conv5x5.weight
        )

        combined_bias = (
            self.alpha2.squeeze() * self.conv1x1.bias + self.alpha3.squeeze() * self.conv3x3.bias + self.alpha4.squeeze() * self.conv5x5.bias
        )

        device = self.conv5x5_reparam.weight.device

        combined_weight = combined_weight.to(device)
        combined_bias = combined_bias.to(device)

        self.conv5x5_reparam.weight = nn.Parameter(combined_weight)
        self.conv5x5_reparam.bias = nn.Parameter(combined_bias)

    def train(self, mode: bool = True):
        super().train(mode)
        if not mode:
            self.reparam_5x5()
        return self

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            out = self.forward_train(x)
        else:
            out = self.conv5x5_reparam(x)
        return out


class Shift(nn.Module):
    def __init__(self, shift: int = 4) -> None:
        super().__init__()
        self.shift = shift

    def forward(self, x: Tensor) -> Tensor:
        return torch.roll(x, shifts=(self.shift, self.shift), dims=(2, 3))


class HybridAttention(nn.Module):
    def __init__(self, dim: int = 64, down: int = 8, shift: int = 0, window_size: int = 8) -> None:
        super().__init__()
        self.att = self.att = nn.Sequential(
            nn.MaxPool2d(down, down) if down > 1 else nn.Identity(),
            Shift(-shift) if shift else nn.Identity(),
            FocusedLinearAttention(dim // 2, window_size),
            Shift(shift) if shift else nn.Identity(),
            nn.Upsample(scale_factor=down, mode='bilinear') if down > 1 else nn.Identity(),
        )
        self.conv = OmniShift(dim // 2)
        self.aggr = nn.Sequential(nn.Conv2d(dim, dim, 1), nn.Mish(True))

    def forward(self, x: Tensor) -> Tensor:
        x1, x2 = x.chunk(2, dim=1)
        x1 = self.conv(x1)
        x2 = self.att(x2)
        return self.aggr(torch.cat([x1, x2], dim=1)) * x


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(
        self,
        dim: int = 64,
        expansion_ratio: float = 1.5,
        conv_ratio: float = 1.0,
        down: int = 1,
        shift: int = 0,
        window_size: int = 8,
    ) -> None:
        super().__init__()
        self.norm = LayerNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Conv2d(dim, hidden * 2, 3, 1, 1)
        self.act = nn.Mish()
        conv_channels = int(conv_ratio * dim)
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]

        self.conv = HybridAttention(dim, down, shift, window_size)
        self.fc2 = nn.Conv2d(hidden, dim, 3, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        shortcut = x
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.conv(c)
        x = self.act(self.fc2(self.act(g) * torch.cat((i, c), dim=1)))
        return x + shortcut


class GatedGroup(nn.Module):
    def __init__(
        self,
        res_blocks: int = 6,
        dim: int = 64,
        down_sample: int = 4,
        expansion_ratio: float = 1.5,
        window_size: int = 8,
    ) -> None:
        super().__init__()
        self.register_buffer('down_sample', torch.tensor(down_sample, dtype=torch.uint8))

        self.body = nn.Sequential(
            *[
                GatedCNNBlock(
                    dim,
                    down=down_sample,
                    expansion_ratio=expansion_ratio,
                    shift=0 if (i % 2 == 0) else window_size // 2,
                    window_size=window_size,
                )
                for i in range(res_blocks)
            ]
            + [OmniShift(dim), nn.Conv2d(dim, dim, kernel_size=1)]
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.body(x) + x


class RHA(nn.Module):
    """Residual Hybrid Attention"""

    def __init__(
        self,
        dim: int = 64,
        scale: int = 4,
        in_ch: int = 3,
        out_ch: int = 3,
        mid_dim: int = 32,
        down_list: Sequence[int] = (
            8,
            4,
        ),
        expansion_ratio: float = 1.5,
        group_blocks: int = 4,
        res_blocks: int = 6,
        upsample: SampleMods = 'pixelshuffledirect',
        unshuffle_mod: bool = False,
        window_size: int = 8,
    ) -> None:
        super().__init__()
        unshuffle = 0
        if scale < 4 and unshuffle_mod:
            if scale == 3:
                raise ValueError('Unshuffle_mod does not support 3x')
            unshuffle = 4 // scale
            self.register_buffer('unshuffle', torch.tensor(unshuffle, dtype=torch.uint8))
            scale = 4

        self.pad: int = unshuffle if unshuffle > 0 else 1
        self.pad *= np.max(down_list) * window_size
        self.to_feat = (
            nn.Sequential(
                nn.PixelUnshuffle(unshuffle),
                nn.Conv2d(in_ch * unshuffle**2, dim, 3, 1, 1),
            )
            if unshuffle_mod
            else (nn.Conv2d(in_ch, dim, 3, 1, 1))
        )

        down_sample_len = len(down_list)
        self.body = nn.Sequential(
            *[
                GatedGroup(
                    res_blocks,
                    dim,
                    down_list[i % down_sample_len],
                    expansion_ratio,
                    window_size,
                )
                for i in range(group_blocks)
            ]
        )
        self.scale = scale
        self.to_img = UniUpsample(upsample, scale, dim, out_ch, mid_dim)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def load_state_dict(self, state_dict, *args, **kwargs):
        state_dict['to_img.MetaUpsample'] = self.to_img.MetaUpsample
        if 'unshuffle' in state_dict:
            state_dict['unshuffle'] = self.unshuffle
        return super().load_state_dict(state_dict, *args, **kwargs)

    def check_img_size(self, x: Tensor, resolution: tuple[int, int]) -> Tensor:
        scaled_size = self.pad
        mod_pad_h = (scaled_size - resolution[0] % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - resolution[1] % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, x: Tensor) -> Tensor:
        b, c, h, w = x.shape
        x = self.check_img_size(x, (h, w))
        x = self.to_feat(x)
        x = self.body(x) + x
        return self.to_img(x)[:, :, : h * self.scale, : w * self.scale]



import math
from typing import Literal

import torch
from torch import nn
from torch.nn.init import trunc_normal_
from torch.nn import functional as F

from ...utilities import DySample

SampleMods = Literal['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']


class UniUpsample(nn.Sequential):
    def __init__(
        self,
        upsample: SampleMods,
        scale: int = 2,
        in_dim: int = 64,
        out_dim: int = 3,
        mid_dim: int = 64,  # Only pixelshuffle
        group: int = 4,  # Only DySample
    ):
        m = []
        if scale == 1 or upsample == 'conv':
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'pixelshuffledirect':
            m.extend([nn.Conv2d(in_dim, out_dim * scale**2, 3, 1, 1), nn.PixelShuffle(scale)])
        elif upsample == 'pixelshuffle':
            m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            if (scale & (scale - 1)) == 0:  # scale = 2^n
                for _ in range(int(math.log2(scale))):
                    m.extend([nn.Conv2d(mid_dim, 4 * mid_dim, 3, 1, 1), nn.PixelShuffle(2)])
            elif scale == 3:
                m.extend([nn.Conv2d(mid_dim, 9 * mid_dim, 3, 1, 1), nn.PixelShuffle(3)])
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'nearest+conv':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        (
                            nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                            nn.Upsample(scale_factor=2),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        )
                    )
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            elif scale == 3:
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.Upsample(scale_factor=scale),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'dysample':
            m.append(DySample(in_dim, out_dim, scale, group))
        else:
            raise ValueError(f'An invalid Upsample was selected. Please choose one of {SampleMods}')
        super().__init__(*m)

        self.register_buffer(
            'MetaUpsample',
            torch.tensor(
                [
                    1,  # Block version, if you change something, please number from the end so that you can distinguish between authorized changes and third parties
                    list(SampleMods.__args__).index(upsample),  # UpSample method index
                    scale,
                    in_dim,
                    out_dim,
                    mid_dim,
                    group,
                ],
                dtype=torch.uint8,
            ),
        )


class LayerNorm(nn.Module):
    def __init__(self, dim: int = 64, eps: float = 1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = eps

    def forward(self, x):
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight[:, None, None] * x + self.bias[:, None, None]


class InceptionDWConv2d(nn.Module):
    """Inception depthweise convolution"""

    def __init__(self, in_channels, square_kernel_size=3, band_kernel_size=11, branch_ratio=0.125):
        super().__init__()

        gc = int(in_channels * branch_ratio)  # channel numbers of a convolution branch
        self.dwconv_hw = nn.Conv2d(gc, gc, square_kernel_size, padding=square_kernel_size // 2, groups=gc)
        self.dwconv_w = nn.Conv2d(gc, gc, kernel_size=(1, band_kernel_size), padding=(0, band_kernel_size // 2), groups=gc)
        self.dwconv_h = nn.Conv2d(gc, gc, kernel_size=(band_kernel_size, 1), padding=(band_kernel_size // 2, 0), groups=gc)
        self.split_indexes = (in_channels - 3 * gc, gc, gc, gc)

    def forward(self, x):
        x_id, x_hw, x_w, x_h = torch.split(x, self.split_indexes, dim=1)
        return torch.cat(
            (x_id, self.dwconv_hw(x_hw), self.dwconv_w(x_w), self.dwconv_h(x_h)),
            dim=1,
        )


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(
        self,
        dim: int = 64,
        expansion_ratio: float = 8 / 3,
        conv_ratio: float = 1.0,
    ):
        super().__init__()
        self.norm = LayerNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Conv2d(dim, hidden * 2, 3, 1, 1)

        self.act = nn.Mish()
        conv_channels = int(conv_ratio * dim)
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]

        self.conv = InceptionDWConv2d(conv_channels)
        self.fc2 = nn.Conv2d(hidden, dim, 3, 1, 1)
        self.gamma = nn.Parameter(torch.ones([1, dim, 1, 1]), requires_grad=True)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m):
        if isinstance(m, nn.Conv2d | nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        shortcut = x
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.conv(c)
        x = self.act(self.fc2(self.act(g) * torch.cat((i, c), dim=1)))
        return (x * self.gamma) + shortcut


class MSG(nn.Module):
    def __init__(self, dim, expansion_msg=1.5):
        super().__init__()
        self.down = nn.Sequential(nn.Conv2d(dim, dim // 4, 3, 1, 1), nn.PixelUnshuffle(2), nn.LeakyReLU(0.1, True))
        self.gated = nn.Sequential(*[GatedCNNBlock(dim, expansion_ratio=expansion_msg) for _ in range(3)])
        self.up = nn.Sequential(nn.Conv2d(dim, dim * 4, 3, 1, 1), nn.PixelShuffle(2), nn.LeakyReLU(0.1, True))

    def forward(self, x):
        out = self.down(x)
        out = self.gated(out)
        return self.up(out) + x


class Blocks(nn.Module):
    def __init__(self, dim: int = 64, blocks: int = 4, expansion_factor: float = 1.5, expansion_msg: float = 1.5):
        super().__init__()
        self.blocks = nn.Sequential(*[GatedCNNBlock(dim, expansion_factor) for _ in range(blocks)])
        self.msg = MSG(dim, expansion_msg)

    def forward(self, x):
        return self.msg(self.blocks(x))


class MoESR(nn.Module):
    """Mamba out Excitation Super-Resolution"""

    def __init__(
        self,
        in_ch: int = 3,
        out_ch: int = 3,
        scale: int = 4,
        dim: int = 64,
        n_blocks: int = 9,
        n_block: int = 4,
        expansion_factor: int = 8 / 3,
        expansion_msg: int = 1.5,
        upsampler: SampleMods = 'pixelshuffledirect',
        upsample_dim: int = 64,
    ):
        super().__init__()
        if upsampler == 'conv':
            scale = 1
        self.scale = scale
        self.in_to_dim = nn.Conv2d(in_ch, dim, 3, 1, 1)
        self.blocks = nn.Sequential(*[Blocks(dim, n_block, expansion_factor, expansion_msg) for i in range(n_blocks)])
        self.upscale = UniUpsample(upsampler, scale, dim, out_ch, upsample_dim)

    def check_img_size(self, x, resolution):
        h, w = resolution
        scaled_size = 2
        mod_pad_h = (scaled_size - h % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - w % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, x):
        b, c, h, w = x.shape
        x = self.check_img_size(x, (h, w))
        x = self.in_to_dim(x)
        x = self.blocks(x) + x
        x = self.upscale(x)[:, :, : h * self.scale, : w * self.scale]
        return x



import math
from typing import Mapping

from .arch import RTMoSR
from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import get_seq_len


class RTMoSRArch(Architecture[RTMoSR]):
    def __init__(self):
        super().__init__(
            uid='RTMoSR',
            detect=KeyCondition.has_all(
                'body.0.norm.scale',
                'body.0.norm.offset',
                'body.0.fc1.alpha',
                'body.0.fc1.conv1.k0',
                'body.0.fc1.conv1.b0',
                'body.0.fc1.conv1.k1',
                'body.0.fc1.conv1.b1',
                'body.0.fc1.conv2.weight',
                'body.0.fc1.conv2.bias',
                'body.0.fc1.conv3.sk.weight',
                'body.0.fc1.conv3.sk.bias',
                'body.0.fc1.conv3.conv.0.weight',
                'body.0.fc1.conv3.conv.0.bias',
                'body.0.fc1.conv3.conv.1.weight',
                'body.0.fc1.conv3.conv.1.bias',
                'body.0.fc1.conv3.conv.2.weight',
                'body.0.fc1.conv3.conv.2.bias',
                'body.0.fc1.conv3.eval_conv.weight',
                'body.0.fc1.conv3.eval_conv.bias',
                'body.0.fc1.conv_3x3_rep.weight',
                'body.0.fc1.conv_3x3_rep.bias',
                'body.0.conv.0.poll.1.alpha',
                'body.0.conv.0.poll.1.conv1.k0',
                'body.0.conv.0.poll.1.conv1.b0',
                'body.0.conv.0.poll.1.conv1.k1',
                'body.0.conv.0.poll.1.conv1.b1',
                'body.0.conv.0.poll.1.conv2.weight',
                'body.0.conv.0.poll.1.conv2.bias',
                'body.0.conv.0.poll.1.conv3.sk.weight',
                'body.0.conv.0.poll.1.conv3.sk.bias',
                'body.0.conv.0.poll.1.conv3.conv.0.weight',
                'body.0.conv.0.poll.1.conv3.conv.0.bias',
                'body.0.conv.0.poll.1.conv3.conv.1.weight',
                'body.0.conv.0.poll.1.conv3.conv.1.bias',
                'body.0.conv.0.poll.1.conv3.conv.2.weight',
                'body.0.conv.0.poll.1.conv3.conv.2.bias',
                'body.0.conv.0.poll.1.conv3.eval_conv.weight',
                'body.0.conv.0.poll.1.conv3.eval_conv.bias',
                'body.0.conv.0.poll.1.conv_3x3_rep.weight',
                'body.0.conv.0.poll.1.conv_3x3_rep.bias',
                'body.0.conv.1.alpha1',
                'body.0.conv.1.alpha2',
                'body.0.conv.1.alpha3',
                'body.0.conv.1.alpha4',
                'body.0.conv.1.conv1x1.weight',
                'body.0.conv.1.conv1x1.bias',
                'body.0.conv.1.conv3x3.weight',
                'body.0.conv.1.conv3x3.bias',
                'body.0.conv.1.conv5x5.weight',
                'body.0.conv.1.conv5x5.bias',
                'body.0.conv.1.conv5x5_reparam.weight',
                'body.0.conv.1.conv5x5_reparam.bias',
                'to_img.0.alpha',
                'to_img.0.conv1.k0',
                'to_img.0.conv1.b0',
                'to_img.0.conv1.k1',
                'to_img.0.conv1.b1',
                'to_img.0.conv2.weight',
                'to_img.0.conv2.bias',
                'to_img.0.conv3.sk.weight',
                'to_img.0.conv3.sk.bias',
                'to_img.0.conv3.conv.0.weight',
                'to_img.0.conv3.conv.0.bias',
                'to_img.0.conv3.conv.1.weight',
                'to_img.0.conv3.conv.1.bias',
                'to_img.0.conv3.conv.2.weight',
                'to_img.0.conv3.conv.2.bias',
                'to_img.0.conv3.eval_conv.weight',
                'to_img.0.conv3.eval_conv.bias',
                'to_img.0.conv_3x3_rep.weight',
                'to_img.0.conv_3x3_rep.bias',
            ),
        )

    def load(self, state: Mapping[str, object]):
        unshuffle = False
        if 'to_feat.1.alpha' in state:
            unshuffle = True
            scale = math.isqrt(state['to_feat.1.conv_3x3_rep.weight'].shape[1] // 3)
            dim = state['to_feat.1.conv_3x3_rep.weight'].shape[0]
        else:
            scale = math.isqrt(state['to_img.0.conv_3x3_rep.weight'].shape[0] // 3)
            dim = state['to_feat.conv_3x3_rep.weight'].shape[0]
        dccm = 'body.0.fc2.alpha' in state
        se = 'body.0.conv.2.squeezing.0.weight' in state
        ffn = state['body.0.fc1.conv_3x3_rep.weight'].shape[0] / dim / 2
        n_blocks = get_seq_len(state, 'body')

        model = RTMoSR(scale=scale, dim=dim, ffn_expansion=ffn, n_blocks=n_blocks, unshuffle_mod=unshuffle, dccm=dccm, se=se)

        return self._enhance_model(model=model, in_channels=3, out_channels=3, upscale=int(2), name='RTMoSR')



import math
from typing import Mapping

from .arch import MoSRv2
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len


class MoSRv2Arch(Architecture[MoSRv2]):
    def __init__(self):
        super().__init__(
            uid='MoSRv2',
            detect=KeyCondition.has_any(
                KeyCondition.has_all(
                    'gblocks.1.weight',
                    'gblocks.1.bias',
                    'gblocks.2.gamma',
                    KeyCondition.has_any(
                        KeyCondition.has_all('gblocks.2.norm.scale', 'gblocks.2.norm.offset'),
                        KeyCondition.has_all('gblocks.2.norm.weight', 'gblocks.2.norm.bias'),
                    ),
                    'gblocks.2.fc1.weight',
                    'gblocks.2.fc1.bias',
                    'gblocks.2.conv.dwconv_hw.weight',
                    'gblocks.2.conv.dwconv_hw.bias',
                    'gblocks.2.conv.dwconv_w.weight',
                    'gblocks.2.conv.dwconv_w.bias',
                    'gblocks.2.conv.dwconv_h.weight',
                    'gblocks.2.conv.dwconv_h.bias',
                    'gblocks.2.fc2.weight',
                    'gblocks.2.fc2.bias',
                    'to_img.MetaUpsample',
                    'to_img.0.weight',
                    'to_img.0.bias',
                ),
                KeyCondition.has_all(
                    'gblocks.0.weight',
                    'gblocks.0.bias',
                    'gblocks.1.gamma',
                    KeyCondition.has_any(
                        KeyCondition.has_all('gblocks.1.norm.scale', 'gblocks.1.norm.offset'),
                        KeyCondition.has_all('gblocks.1.norm.weight', 'gblocks.1.norm.bias'),
                    ),
                    'gblocks.1.fc1.weight',
                    'gblocks.1.fc1.bias',
                    'gblocks.1.conv.dwconv_hw.weight',
                    'gblocks.1.conv.dwconv_hw.bias',
                    'gblocks.1.conv.dwconv_w.weight',
                    'gblocks.1.conv.dwconv_w.bias',
                    'gblocks.1.conv.dwconv_h.weight',
                    'gblocks.1.conv.dwconv_h.bias',
                    'gblocks.1.fc2.weight',
                    'gblocks.1.fc2.bias',
                    'to_img.MetaUpsample',
                    'to_img.0.weight',
                    'to_img.0.bias',
                ),
            ),
        )

    def load(self, state: Mapping[str, object]):
        samplemods = ['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']
        _, upsampler, scale, dim, in_ch, mid_dim, _ = [i.item() for i in state['to_img.MetaUpsample']]
        upsampler = samplemods[upsampler]
        n_block = get_seq_len(state, 'gblocks')
        if 'gblocks.0.weight' in state:
            unshuffle_mod = False
            n_block -= 6
            expansion_ratio = state['gblocks.1.fc1.weight'].shape[0] // 2 / dim
            rms_norm = 'gblocks.1.norm.scale' in state
        else:
            scale = math.isqrt(state['gblocks.1.weight'].shape[1] // in_ch)
            n_block -= 7
            unshuffle_mod = True
            expansion_ratio = state['gblocks.2.fc1.weight'].shape[0] // 2 / dim
            rms_norm = 'gblocks.2.norm.scale' in state

        model = MoSRv2(
            in_ch=in_ch,
            scale=scale,
            n_block=n_block,
            dim=dim,
            upsampler=upsampler,
            expansion_ratio=expansion_ratio,
            mid_dim=mid_dim,
            unshuffle_mod=unshuffle_mod,
            rms_norm=rms_norm,
        )

        return self._enhance_model(model=model, in_channels=in_ch, out_channels=in_ch, upscale=scale, name='MoSRv2')



from typing import Mapping

from .arch import GateRV2
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len


class GateRV2Arch(Architecture[GateRV2]):
    def __init__(self):
        super().__init__(
            uid='GateRv2',
            detect=KeyCondition.has_all(
                'in_to_dim.weight',
                'in_to_dim.bias',
                'encode.0.gated.0.gamma0',
                'encode.0.gated.0.gamma1',
                'encode.0.gated.0.local.0.scale',
                'encode.0.gated.0.local.0.offset',
                'encode.0.gated.0.local.1.weight',
                'encode.0.gated.0.local.1.bias',
                'encode.0.gated.0.local.2.weight',
                'encode.0.gated.0.local.2.bias',
                'encode.0.gated.0.sca.1.weight',
                'encode.0.gated.0.sca.1.bias',
                'encode.0.gated.0.glob.norm.scale',
                'encode.0.gated.0.glob.norm.offset',
                'encode.0.gated.0.glob.fc1.weight',
                'encode.0.gated.0.glob.fc1.bias',
                'encode.0.gated.0.glob.token_mix.dwconv_hw.weight',
                'encode.0.gated.0.glob.token_mix.dwconv_hw.bias',
                'encode.0.gated.0.glob.token_mix.dwconv_w.weight',
                'encode.0.gated.0.glob.token_mix.dwconv_w.bias',
                'encode.0.gated.0.glob.token_mix.dwconv_h.weight',
                'encode.0.gated.0.glob.token_mix.dwconv_h.bias',
                'encode.0.gated.0.glob.fc2.weight',
                'encode.0.gated.0.glob.fc2.bias',
                'encode.0.scale.0.weight',
                'encode.1.gated.0.gamma0',
                'encode.1.gated.0.gamma1',
                'encode.1.gated.0.local.0.scale',
                'encode.1.gated.0.local.0.offset',
                'encode.1.gated.0.local.1.weight',
                'encode.1.gated.0.local.1.bias',
                'encode.1.gated.0.local.2.weight',
                'encode.1.gated.0.local.2.bias',
                'encode.1.gated.0.sca.1.weight',
                'encode.1.gated.0.sca.1.bias',
                'encode.1.gated.0.glob.norm.scale',
                'encode.1.gated.0.glob.norm.offset',
                'encode.1.gated.0.glob.fc1.weight',
                'encode.1.gated.0.glob.fc1.bias',
                'encode.1.gated.0.glob.token_mix.dwconv_hw.weight',
                'encode.1.gated.0.glob.token_mix.dwconv_hw.bias',
                'encode.1.gated.0.glob.token_mix.dwconv_w.weight',
                'encode.1.gated.0.glob.token_mix.dwconv_w.bias',
                'encode.1.gated.0.glob.token_mix.dwconv_h.weight',
                'encode.1.gated.0.glob.token_mix.dwconv_h.bias',
                'encode.1.gated.0.glob.fc2.weight',
                'encode.1.gated.0.glob.fc2.bias',
                'encode.1.scale.0.weight',
                'latent.0.norm.scale',
                'latent.0.norm.offset',
                'latent.0.fc1.weight',
                'latent.0.fc1.bias',
                'latent.0.token_mix.query_conv.weight',
                'latent.0.token_mix.query_conv.bias',
                'latent.0.token_mix.key_conv.weight',
                'latent.0.token_mix.key_conv.bias',
                'latent.0.token_mix.value_conv.weight',
                'latent.0.token_mix.value_conv.bias',
                'latent.0.fc2.weight',
                'latent.0.fc2.bias',
                'decode.0.scale.0.weight',
                'decode.0.gated.0.gamma0',
                'decode.0.gated.0.gamma1',
                'decode.0.gated.0.local.0.scale',
                'decode.0.gated.0.local.0.offset',
                'decode.0.gated.0.local.1.weight',
                'decode.0.gated.0.local.1.bias',
                'decode.0.gated.0.local.2.weight',
                'decode.0.gated.0.local.2.bias',
                'decode.0.gated.0.sca.1.weight',
                'decode.0.gated.0.sca.1.bias',
                'decode.0.gated.0.glob.norm.scale',
                'decode.0.gated.0.glob.norm.offset',
                'decode.0.gated.0.glob.fc1.weight',
                'decode.0.gated.0.glob.fc1.bias',
                'decode.0.gated.0.glob.token_mix.dwconv_hw.weight',
                'decode.0.gated.0.glob.token_mix.dwconv_hw.bias',
                'decode.0.gated.0.glob.token_mix.dwconv_w.weight',
                'decode.0.gated.0.glob.token_mix.dwconv_w.bias',
                'decode.0.gated.0.glob.token_mix.dwconv_h.weight',
                'decode.0.gated.0.glob.token_mix.dwconv_h.bias',
                'decode.0.gated.0.glob.fc2.weight',
                'decode.0.gated.0.glob.fc2.bias',
                'decode.0.shor.weight',
                'decode.0.shor.bias',
                'decode.1.scale.0.weight',
                'decode.1.gated.0.gamma0',
                'decode.1.gated.0.gamma1',
                'decode.1.gated.0.local.0.scale',
                'decode.1.gated.0.local.0.offset',
                'decode.1.gated.0.local.1.weight',
                'decode.1.gated.0.local.1.bias',
                'decode.1.gated.0.local.2.weight',
                'decode.1.gated.0.local.2.bias',
                'decode.1.gated.0.sca.1.weight',
                'decode.1.gated.0.sca.1.bias',
                'decode.1.gated.0.glob.norm.scale',
                'decode.1.gated.0.glob.norm.offset',
                'decode.1.gated.0.glob.fc1.weight',
                'decode.1.gated.0.glob.fc1.bias',
                'decode.1.gated.0.glob.token_mix.dwconv_hw.weight',
                'decode.1.gated.0.glob.token_mix.dwconv_hw.bias',
                'decode.1.gated.0.glob.token_mix.dwconv_w.weight',
                'decode.1.gated.0.glob.token_mix.dwconv_w.bias',
                'decode.1.gated.0.glob.token_mix.dwconv_h.weight',
                'decode.1.gated.0.glob.token_mix.dwconv_h.bias',
                'decode.1.gated.0.glob.fc2.weight',
                'decode.1.gated.0.glob.fc2.bias',
                'decode.1.shor.weight',
                'decode.1.shor.bias',
            ),
        )

    def load(self, state: Mapping[str, object]):
        dim, in_ch = state['in_to_dim.weight'].shape[:2]
        enc_blocks = [get_seq_len(state, f'encode.{i}.gated') for i in range(get_seq_len(state, 'encode'))]
        latent = get_seq_len(state, 'latent')
        dec_blocks = [get_seq_len(state, f'decode.{i}.gated') for i in range(get_seq_len(state, 'decode'))]
        if 'upsample.MetaUpsample' in state:
            upsample = ['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']
            _, index, scale, _, out_ch, upsample_dim, _ = [value.item() for value in state['to_img.MetaUpsample']]
            upsampler = upsample[int(index)]
        else:
            scale, upsample_dim, upsampler = 1, 32, 'conv'
        model = GateRV2(in_ch, dim, enc_blocks, dec_blocks, latent, scale, upsampler, upsample_dim)

        return self._enhance_model(model=model, in_channels=in_ch, out_channels=int(in_ch), upscale=scale, name='GateRv2')



import math
from typing import Mapping

from .arch import DRCT
from ...factory import Architecture, KeyCondition
from ...utilities.state_dict import get_seq_len, get_pixelshuffle_params


class MoSRArch(Architecture[DRCT]):
    def __init__(self):
        super().__init__(
            uid='DRCT',
            detect=KeyCondition.has_all(
                'conv_first.weight',
                'conv_first.bias',
                'layers.0.swin1.norm1.weight',
                'layers.0.swin1.norm1.bias',
                'layers.0.swin1.attn.relative_position_bias_table',
                'layers.0.swin1.attn.relative_position_index',
                'layers.0.swin1.attn.qkv.weight',
                'layers.0.swin1.attn.proj.weight',
                'layers.0.swin1.attn.proj.bias',
                'layers.0.swin1.norm2.weight',
                'layers.0.swin1.mlp.fc1.weight',
                'layers.0.swin1.mlp.fc1.bias',
                'layers.0.swin1.mlp.fc2.weight',
                'layers.0.adjust1.weight',
                'layers.0.swin2.norm1.weight',
                'layers.0.adjust2.weight',
                'layers.0.swin3.norm1.weight',
                'layers.0.adjust3.weight',
                'layers.0.swin4.norm1.weight',
                'layers.0.adjust4.weight',
                'layers.0.swin5.norm1.weight',
                'layers.0.adjust5.weight',
                'norm.weight',
                'norm.bias',
            ),
        )

    def load(self, state_dict: Mapping[str, object]):
        patch_size = 1  # cannot be detected
        img_range = 1.0  # cannot be deduced from state_dict

        # detect
        in_chans = state_dict['conv_first.weight'].shape[1]
        embed_dim = state_dict['conv_first.weight'].shape[0]

        num_layers = get_seq_len(state_dict, 'layers')
        depths = (6,) * num_layers
        num_heads = []
        for i in range(num_layers):
            num_heads.append(state_dict[f'layers.{i}.swin1.attn.relative_position_bias_table'].shape[1])

        mlp_ratio = state_dict['layers.0.swin1.mlp.fc1.weight'].shape[0] / embed_dim

        window_square = state_dict['layers.0.swin1.attn.relative_position_bias_table'].shape[0]
        window_size = (math.isqrt(window_square) + 1) // 2

        if 'conv_last.weight' in state_dict:
            upsampler = 'pixelshuffle'
            upscale, _ = get_pixelshuffle_params(state_dict, 'upsample')
        else:
            upsampler = ''
            upscale = 1

        if 'conv_after_body.weight' in state_dict:
            resi_connection = '1conv'
        else:
            resi_connection = 'identity'

        qkv_bias = 'layers.0.swin1.attn.qkv.bias' in state_dict
        gc = state_dict['layers.0.adjust1.weight'].shape[0]

        patch_norm = 'patch_embed.norm.weight' in state_dict
        ape = 'absolute_pos_embed' in state_dict

        if 'layers.0.swin2.attn_mask' in state_dict:
            img_size = math.isqrt(state_dict['layers.0.swin2.attn_mask'].shape[0]) * window_size * patch_size
        else:
            # we only know that the input size is <= window_size,
            # so we just assume that the input size is window_size
            img_size = window_size * patch_size

        model = DRCT(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
            depths=depths,
            num_heads=num_heads,
            window_size=window_size,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            ape=ape,
            patch_norm=patch_norm,
            upscale=upscale,
            img_range=img_range,
            upsampler=upsampler,
            resi_connection=resi_connection,
            gc=gc,
        )

        return self._enhance_model(model=model, in_channels=in_chans, out_channels=in_chans, upscale=upscale, name='DRCT')



from typing import Mapping

from .arch import GateRV3
from ...factory import KeyCondition, Architecture
from ...utilities.state_dict import get_seq_len


class GateRV3Arch(Architecture[GateRV3]):
    def __init__(self):
        super().__init__(
            uid='GateRV3',
            detect=KeyCondition.has_all(
                'in_to_dim.weight',
                'in_to_dim.bias',
                'gater_encode.0.gated.0.gamma0',
                'gater_encode.0.gated.0.gamma1',
                'gater_encode.0.gated.0.local.0.scale',
                'gater_encode.0.gated.0.local.0.offset',
                'gater_encode.0.gated.0.local.1.weight',
                'gater_encode.0.gated.0.local.1.bias',
                'gater_encode.0.gated.0.local.2.weight',
                'gater_encode.0.gated.0.local.2.bias',
                'gater_encode.0.gated.0.sca.1.weight',
                'gater_encode.0.gated.0.sca.1.bias',
                'gater_encode.0.gated.0.glob.norm.scale',
                'gater_encode.0.gated.0.glob.norm.offset',
                'gater_encode.0.gated.0.glob.fc1.weight',
                'gater_encode.0.gated.0.glob.fc1.bias',
                'gater_encode.0.gated.0.glob.token_mix.dwconv_hw.weight',
                'gater_encode.0.gated.0.glob.token_mix.dwconv_hw.bias',
                'gater_encode.0.gated.0.glob.token_mix.dwconv_w.weight',
                'gater_encode.0.gated.0.glob.token_mix.dwconv_w.bias',
                'gater_encode.0.gated.0.glob.token_mix.dwconv_h.weight',
                'gater_encode.0.gated.0.glob.token_mix.dwconv_h.bias',
                'gater_encode.0.gated.0.glob.fc2.weight',
                'gater_encode.0.gated.0.glob.fc2.bias',
                'gater_encode.0.scale.0.weight',
                'span_block0.c1_r.sk.weight',
                'span_block0.c1_r.conv.0.weight',
                'span_block0.c1_r.conv.1.weight',
                'span_block0.c1_r.conv.2.weight',
                'span_block0.c1_r.eval_conv.weight',
                'span_block0.c2_r.sk.weight',
                'span_block0.c2_r.conv.0.weight',
                'span_block0.c2_r.conv.1.weight',
                'span_block0.c2_r.conv.2.weight',
                'span_block0.c2_r.eval_conv.weight',
                'span_block0.c3_r.sk.weight',
                'span_block0.c3_r.conv.0.weight',
                'span_block0.c3_r.conv.1.weight',
                'span_block0.c3_r.conv.2.weight',
                'span_block0.c3_r.eval_conv.weight',
                'span_n_b.0.c1_r.sk.weight',
                'span_n_b.0.c1_r.conv.0.weight',
                'span_n_b.0.c1_r.conv.1.weight',
                'span_n_b.0.c1_r.conv.2.weight',
                'span_n_b.0.c1_r.eval_conv.weight',
                'span_n_b.0.c2_r.sk.weight',
                'span_n_b.0.c2_r.conv.0.weight',
                'span_n_b.0.c2_r.conv.1.weight',
                'span_n_b.0.c2_r.conv.2.weight',
                'span_n_b.0.c2_r.eval_conv.weight',
                'span_n_b.0.c3_r.sk.weight',
                'span_n_b.0.c3_r.conv.0.weight',
                'span_n_b.0.c3_r.conv.1.weight',
                'span_n_b.0.c3_r.conv.2.weight',
                'span_n_b.0.c3_r.eval_conv.weight',
                'span_end.c1_r.sk.weight',
                'span_end.c1_r.conv.0.weight',
                'span_end.c1_r.conv.1.weight',
                'span_end.c1_r.conv.2.weight',
                'span_end.c1_r.eval_conv.weight',
                'span_end.c2_r.sk.weight',
                'span_end.c2_r.conv.0.weight',
                'span_end.c2_r.conv.1.weight',
                'span_end.c2_r.conv.2.weight',
                'span_end.c2_r.eval_conv.weight',
                'span_end.c3_r.sk.weight',
                'span_end.c3_r.conv.0.weight',
                'span_end.c3_r.conv.1.weight',
                'span_end.c3_r.conv.2.weight',
                'span_end.c3_r.eval_conv.weight',
                'sisr_end_conv.sk.weight',
                'sisr_end_conv.sk.bias',
                'sisr_end_conv.conv.0.weight',
                'sisr_end_conv.conv.0.bias',
                'sisr_end_conv.conv.1.weight',
                'sisr_end_conv.conv.1.bias',
                'sisr_end_conv.conv.2.weight',
                'sisr_end_conv.conv.2.bias',
                'sisr_end_conv.eval_conv.weight',
                'sisr_end_conv.eval_conv.bias',
                'sisr_cat_conv.weight',
                'sisr_cat_conv.bias',
                'decode.0.scale.0.weight',
                'decode.0.gated.0.gamma0',
                'decode.0.gated.0.gamma1',
                'decode.0.gated.0.local.0.scale',
                'decode.0.gated.0.local.0.offset',
                'decode.0.gated.0.local.1.weight',
                'decode.0.gated.0.local.1.bias',
                'decode.0.gated.0.local.2.weight',
                'decode.0.gated.0.local.2.bias',
                'decode.0.gated.0.sca.1.weight',
                'decode.0.gated.0.sca.1.bias',
                'decode.0.gated.0.glob.norm.scale',
                'decode.0.gated.0.glob.norm.offset',
                'decode.0.gated.0.glob.fc1.weight',
                'decode.0.gated.0.glob.fc1.bias',
                'decode.0.gated.0.glob.token_mix.dwconv_hw.weight',
                'decode.0.gated.0.glob.token_mix.dwconv_hw.bias',
                'decode.0.gated.0.glob.token_mix.dwconv_w.weight',
                'decode.0.gated.0.glob.token_mix.dwconv_w.bias',
                'decode.0.gated.0.glob.token_mix.dwconv_h.weight',
                'decode.0.gated.0.glob.token_mix.dwconv_h.bias',
                'decode.0.gated.0.glob.fc2.weight',
                'decode.0.gated.0.glob.fc2.bias',
                'decode.0.shor.weight',
                'decode.0.shor.bias',
            ),
        )

    def load(self, state: Mapping[str, object]):
        dim, in_ch = state['in_to_dim.weight'].shape[:2]
        enc_blocks = [get_seq_len(state, f'gater_encode.{i}.gated') for i in range(get_seq_len(state, 'gater_encode'))]
        latent = get_seq_len(state, 'latent')
        dec_blocks = [get_seq_len(state, f'decode.{i}.gated') for i in range(get_seq_len(state, 'decode'))]
        end_kernel = 1
        if 'dim_to_in.MetaUpsample' in state:
            upsample = ['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample', 'transpose+conv', 'lda', 'pa_up']
            _, index, scale, _, out_ch, upsample_dim, _ = [value.item() for value in state['dim_to_in.MetaUpsample']]
            upsampler = upsample[int(index)]

            if upsampler == 'dysample' and 'dim_to_in.0.weight' not in state:
                upsample_dim = dim
                end_kernel = state['dim_to_in.0.end_conv.weight'].shape[2]
            elif upsampler == 'dysample':
                end_kernel = state['dim_to_in.2.end_conv.weight'].shape[2]
        else:
            scale, upsample_dim, upsampler = 1, 32, 'conv'
        attention = 'latent.0.token_mix.qkv_dwconv.weight' in state
        span_blocks = get_seq_len(state, 'span_n_b')
        model = GateRV3(
            in_ch,
            dim,
            enc_blocks,
            dec_blocks,
            latent,
            scale,
            upsampler,
            upsample_dim,
            attention=attention,
            span_blocks=span_blocks,
            end_kernel=end_kernel,
        )

        return self._enhance_model(model=model, in_channels=in_ch, out_channels=int(in_ch), upscale=scale, name='GateRV3')



import torch
import torch.nn.functional as F


def pad_to_multiple(
    tensor: torch.Tensor,
    multiple: int,
    *,
    mode: str,
    value: float = 0.0,
) -> torch.Tensor:
    """
    Pad a tensor's size to a multiple of a number.

    Args:
        tensor: Tensor to pad.
        multiple: Size multiple to pad to.
        mode: Padding mode; see `torch.nn.functional.pad`.
        value: Padding value; see `torch.nn.functional.pad`.

    Returns:
        Padded tensor, or the original tensor if no padding was needed.
    """
    _, _, h, w = tensor.size()
    pad_h = (multiple - h % multiple) % multiple
    pad_w = (multiple - w % multiple) % multiple
    if pad_h or pad_w:
        return F.pad(tensor, (0, pad_w, 0, pad_h), mode, value)
    return tensor



import torch
from torch import nn
from torch.nn.init import trunc_normal_

from resselt.utilities import DySample


class GPS(nn.Module):
    """Geo ensemble PielShuffle"""

    def __init__(
        self,
        dim,
        scale,
        out_ch=3,
        # Own parameters
        kernel_size: int = 3,
    ):
        super().__init__()
        self.in_to_k = nn.Conv2d(dim, scale * scale * out_ch * 8, kernel_size, 1, kernel_size // 2)
        self.ps = nn.PixelShuffle(scale)

    def forward(self, x):
        rgb = self._geo_ensemble(x)
        rgb = self.ps(rgb)
        return rgb

    def _geo_ensemble(self, x):
        x = self.in_to_k(x)
        x = x.reshape(x.shape[0], 8, -1, x.shape[-2], x.shape[-1])
        x = x.mean(dim=1)
        return x


class LayerNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = eps

    def forward(self, x):
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight[:, None, None] * x + self.bias[:, None, None]


class ConvBlock(nn.Module):
    r"""https://github.com/joshyZhou/AST/blob/main/model.py#L22"""

    def __init__(self, in_channel: int, out_channel: int, strides: int = 1):
        super(ConvBlock, self).__init__()
        self.strides = strides
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.block = nn.Sequential(
            nn.Conv2d(in_channel, out_channel, kernel_size=3, stride=strides, padding=1),
            nn.Mish(),
            nn.Conv2d(out_channel, out_channel, kernel_size=3, stride=strides, padding=1),
            nn.Mish(),
        )
        self.conv11 = nn.Conv2d(in_channel, out_channel, kernel_size=1, stride=strides, padding=0)

    def forward(self, x):
        out1 = self.block(x)
        out2 = self.conv11(x)
        out = out1 + out2
        return out


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(self, dim: int, expansion_ratio: float = 8 / 3, conv_ratio: float = 1.0, kernel_size: int = 7):
        super().__init__()
        self.norm = LayerNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Conv2d(dim, hidden * 2, 3, 1, 1)

        self.act = nn.Mish()
        conv_channels = int(conv_ratio * dim)
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]

        self.conv = nn.Conv2d(conv_channels, conv_channels, kernel_size, 1, kernel_size // 2, groups=conv_channels)
        self.fc2 = nn.Conv2d(hidden, dim, 3, 1, 1)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m):
        if isinstance(m, nn.Conv2d | nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        shortcut = x
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.conv(c)
        x = self.act(self.fc2(self.act(g) * torch.cat((i, c), dim=1)))
        return x + (shortcut - 0.5)


class mosr(nn.Module):
    """Mamba Out Super-Resolution"""

    def __init__(
        self,
        in_ch: int = 3,
        out_ch: int = 3,
        upscale: int = 4,
        n_block: int = 24,
        dim: int = 64,
        upsampler: str = 'ps',  # "ps" "ds"
        kernel_size: int = 7,
        expansion_ratio: float = 1.5,
        conv_ratio: float = 1.0,
    ):
        super(mosr, self).__init__()
        if upsampler in ['ps']:
            out_ch = in_ch
        self.gblocks = nn.Sequential(
            *[nn.Conv2d(in_ch, dim, 3, 1, 1)]
            + [
                GatedCNNBlock(
                    dim=dim,
                    expansion_ratio=expansion_ratio,
                    kernel_size=kernel_size,
                    conv_ratio=conv_ratio,
                )
                for _ in range(n_block)
            ]
            + [nn.Conv2d(dim, dim * 2, 3, 1, 1), nn.Mish(), nn.Conv2d(dim * 2, dim, 3, 1, 1), nn.Mish(), nn.Conv2d(dim, dim, 1, 1)]
        )

        self.shortcut = ConvBlock(in_ch, dim)

        if upsampler == 'ps':
            self.upsampler = nn.Sequential(nn.Conv2d(dim, out_ch * (upscale**2), 3, 1, 1), nn.PixelShuffle(upscale))
        elif upsampler == 'dys':
            self.upsampler = DySample(dim, out_ch, upscale)
        elif upsampler == 'gps':
            self.upsampler = GPS(dim, upscale, out_ch)
        else:
            raise NotImplementedError(
                f'upsampler: {upsampler} not supported, choose one of these options: \
                ["ps", "dys"]'
            )

    def forward(self, x):
        x = self.gblocks(x) + (self.shortcut(x) - 0.5)
        return self.upsampler(x)



import math
from typing import Literal

import torch
import torch.nn.functional as F
from torch import nn
from torch.nn.init import Tensor, trunc_normal_


SampleMods = Literal['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']


class DySample(nn.Module):
    """Adapted from 'Learning to Upsample by Learning to Sample':
    https://arxiv.org/abs/2308.15085
    https://github.com/tiny-smart/dysample
    """

    def __init__(
        self,
        in_channels: int,
        out_ch: int,
        scale: int = 2,
        groups: int = 4,
        end_convolution: bool = True,
    ) -> None:
        super().__init__()

        if in_channels <= groups or in_channels % groups != 0:
            msg = 'Incorrect in_channels and groups values.'
            raise ValueError(msg)

        out_channels = 2 * groups * scale**2
        self.scale = scale
        self.groups = groups
        self.end_convolution = end_convolution
        if end_convolution:
            self.end_conv = nn.Conv2d(in_channels, out_ch, kernel_size=1)

        self.offset = nn.Conv2d(in_channels, out_channels, 1)
        self.scope = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        if self.training:
            nn.init.trunc_normal_(self.offset.weight, std=0.02)
            nn.init.constant_(self.scope.weight, val=0)

        self.register_buffer('init_pos', self._init_pos())

    def _init_pos(self) -> Tensor:
        h = torch.arange((-self.scale + 1) / 2, (self.scale - 1) / 2 + 1) / self.scale
        return torch.stack(torch.meshgrid([h, h], indexing='ij')).transpose(1, 2).repeat(1, self.groups, 1).reshape(1, -1, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        offset = self.offset(x) * self.scope(x).sigmoid() * 0.5 + self.init_pos
        B, _, H, W = offset.shape
        offset = offset.view(B, 2, -1, H, W)
        coords_h = torch.arange(H) + 0.5
        coords_w = torch.arange(W) + 0.5

        coords = (
            torch.stack(torch.meshgrid([coords_w, coords_h], indexing='ij'))
            .transpose(1, 2)
            .unsqueeze(1)
            .unsqueeze(0)
            .type(x.dtype)
            .to(x.device, non_blocking=True)
        )
        normalizer = torch.tensor([W, H], dtype=x.dtype, device=x.device, pin_memory=True).view(1, 2, 1, 1, 1)
        coords = 2 * (coords + offset) / normalizer - 1

        coords = (
            F.pixel_shuffle(coords.reshape(B, -1, H, W), self.scale)
            .view(B, 2, -1, self.scale * H, self.scale * W)
            .permute(0, 2, 3, 4, 1)
            .contiguous()
            .flatten(0, 1)
        )
        output = F.grid_sample(
            x.reshape(B * self.groups, -1, H, W),
            coords,
            mode='bilinear',
            align_corners=False,
            padding_mode='border',
        ).view(B, -1, self.scale * H, self.scale * W)

        if self.end_convolution:
            output = self.end_conv(output)

        return output


class UniUpsample(nn.Sequential):
    def __init__(
        self,
        upsample: SampleMods,
        scale: int = 2,
        in_dim: int = 64,
        out_dim: int = 3,
        mid_dim: int = 64,  # Only pixelshuffle and DySample
        group: int = 4,  # Only DySample
    ) -> None:
        m = []

        if scale == 1 or upsample == 'conv':
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'pixelshuffledirect':
            m.extend([nn.Conv2d(in_dim, out_dim * scale**2, 3, 1, 1), nn.PixelShuffle(scale)])
        elif upsample == 'pixelshuffle':
            m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            if (scale & (scale - 1)) == 0:  # scale = 2^n
                for _ in range(int(math.log2(scale))):
                    m.extend([nn.Conv2d(mid_dim, 4 * mid_dim, 3, 1, 1), nn.PixelShuffle(2)])
            elif scale == 3:
                m.extend([nn.Conv2d(mid_dim, 9 * mid_dim, 3, 1, 1), nn.PixelShuffle(3)])
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'nearest+conv':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        (
                            nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                            nn.Upsample(scale_factor=2),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        )
                    )
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            elif scale == 3:
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.Upsample(scale_factor=scale),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'dysample':
            if mid_dim != in_dim:
                m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
                dys_dim = mid_dim
            else:
                dys_dim = in_dim
            m.append(DySample(dys_dim, out_dim, scale, group))
        else:
            raise ValueError(f'An invalid Upsample was selected. Please choose one of {SampleMods}')
        super().__init__(*m)

        self.register_buffer(
            'MetaUpsample',
            torch.tensor(
                [
                    2,  # Block version, if you change something, please number from the end so that you can distinguish between authorized changes and third parties
                    list(SampleMods.__args__).index(upsample),  # UpSample method index
                    scale,
                    in_dim,
                    out_dim,
                    mid_dim,
                    group,
                ],
                dtype=torch.uint8,
            ),
        )


class InceptionDWConv2d(nn.Module):
    """Inception depthweise convolution"""

    def __init__(
        self,
        in_channels: int = 64,
        square_kernel_size: int = 3,
        band_kernel_size: int = 11,
        branch_ratio: int = 0.125,
    ) -> None:
        super().__init__()

        gc = int(in_channels * branch_ratio)  # channel numbers of a convolution branch
        self.dwconv_hw = nn.Conv2d(gc, gc, square_kernel_size, padding=square_kernel_size // 2, groups=gc)
        self.dwconv_w = nn.Conv2d(
            gc,
            gc,
            kernel_size=(1, band_kernel_size),
            padding=(0, band_kernel_size // 2),
            groups=gc,
        )
        self.dwconv_h = nn.Conv2d(
            gc,
            gc,
            kernel_size=(band_kernel_size, 1),
            padding=(band_kernel_size // 2, 0),
            groups=gc,
        )
        self.split_indexes = (in_channels - 3 * gc, gc, gc, gc)

    def forward(self, x: Tensor) -> Tensor:
        x_id, x_hw, x_w, x_h = torch.split(x, self.split_indexes, dim=1)
        return torch.cat(
            (x_id, self.dwconv_hw(x_hw), self.dwconv_w(x_w), self.dwconv_h(x_h)),
            dim=1,
        )


class RMSNorm(nn.Module):
    def __init__(self, dim: int = 64, eps: float = 1e-6) -> None:
        super().__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(dim, 1, 1))
        self.offset = nn.Parameter(torch.zeros(dim, 1, 1))

    def forward(self, x: Tensor) -> Tensor:
        norm_x = x.norm(2, dim=1, keepdim=True)
        d_x = x.size(1)
        rms_x = norm_x * (d_x ** (-1.0 / 2))
        x_normed = x / (rms_x + self.eps)
        return self.scale * x_normed + self.offset


class LayerNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(dim))
        self.bias = nn.Parameter(torch.zeros(dim))
        self.eps = eps
        self.dim = (dim,)

    def forward(self, x):
        if x.is_contiguous(memory_format=torch.channels_last):
            return F.layer_norm(x.permute(0, 2, 3, 1), self.dim, self.weight, self.bias, self.eps).permute(0, 3, 1, 2)
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        return self.weight[:, None, None] * x + self.bias[:, None, None]


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(self, dim: int = 64, expansion_ratio: float = 8 / 3, rms_norm: bool = True) -> None:
        super().__init__()
        self.norm = RMSNorm(dim) if rms_norm else LayerNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Conv2d(dim, hidden * 2, 3, 1, 1)

        self.act = nn.Mish()
        conv_channels = dim
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]

        self.conv = InceptionDWConv2d(conv_channels)
        self.fc2 = nn.Conv2d(hidden, dim, 3, 1, 1)
        self.gamma = nn.Parameter(torch.ones([1, dim, 1, 1]), requires_grad=True)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x: Tensor) -> Tensor:
        shortcut = x
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.conv(c)
        x = self.act(self.fc2(self.act(g) * torch.cat((i, c), dim=1)))
        return x * self.gamma + shortcut


class MoSRv2(nn.Module):
    """Mamba Out Super-Resolution"""

    def __init__(
        self,
        in_ch: int = 3,
        scale: int = 4,
        n_block: int = 24,
        dim: int = 64,
        upsampler: SampleMods = 'pixelshuffledirect',
        expansion_ratio: float = 1.5,
        mid_dim=32,
        unshuffle_mod: bool = True,
        rms_norm: bool = False,
    ) -> None:
        super().__init__()
        self.short = nn.Upsample(scale_factor=scale, mode='bilinear')
        self.scale = scale
        self.pad = 1
        if unshuffle_mod and scale < 3:
            unshuffle = 4 // scale
            in_to_dim = [
                nn.PixelUnshuffle(unshuffle),
                nn.Conv2d(in_ch * unshuffle**2, dim, 3, 1, 1),
            ]
            self.pad = unshuffle
            scale = 4
        else:
            in_to_dim = [nn.Conv2d(in_ch, dim, 3, 1, 1)]

        self.gblocks = nn.Sequential(
            *in_to_dim
            + [GatedCNNBlock(dim=dim, expansion_ratio=expansion_ratio, rms_norm=rms_norm) for _ in range(n_block)]
            + [
                nn.Conv2d(dim, dim * 2, 3, 1, 1),
                nn.Mish(True),
                nn.Conv2d(dim * 2, dim, 3, 1, 1),
                nn.Mish(True),
                nn.Conv2d(dim, dim, 1, 1),
            ]
        )
        self.to_img = UniUpsample(upsampler, scale, dim, in_ch, mid_dim)

    def load_state_dict(self, state_dict, *args, **kwargs):
        state_dict['to_img.MetaUpsample'] = self.to_img.MetaUpsample
        return super().load_state_dict(state_dict, *args, **kwargs)

    def check_img_size(self, x: Tensor, h: int, w: int) -> Tensor:
        mod_pad_h = (self.pad - h % self.pad) % self.pad
        mod_pad_w = (self.pad - w % self.pad) % self.pad
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, x: Tensor) -> Tensor:
        b, c, h, w = x.shape
        x = self.check_img_size(x, h, w)
        x = self.to_img(self.gblocks(x)) + self.short(x)
        return x[:, :, : h * self.scale, : w * self.scale]



import torch
from torch import nn
from torch.nn import functional as F


class DySample(nn.Module):
    """Adapted from 'Learning to Upsample by Learning to Sample':
    https://arxiv.org/abs/2308.15085
    https://github.com/tiny-smart/dysample
    """

    def __init__(
        self,
        in_channels: int,
        out_ch: int,
        scale: int = 2,
        groups: int = 4,
        end_convolution: bool = True,
    ):
        super().__init__()

        try:
            assert in_channels >= groups and in_channels % groups == 0
        except Exception:
            msg = 'Incorrect in_channels and groups values.'
            raise ValueError(msg)

        out_channels = 2 * groups * scale**2
        self.scale = scale
        self.groups = groups
        self.end_convolution = end_convolution
        if end_convolution:
            self.end_conv = nn.Conv2d(in_channels, out_ch, kernel_size=1)

        self.offset = nn.Conv2d(in_channels, out_channels, 1)
        self.scope = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        if self.training:
            nn.init.trunc_normal_(self.offset.weight, std=0.02)
            nn.init.constant_(self.scope.weight, val=0)

        self.register_buffer('init_pos', self._init_pos())

    def _init_pos(self):
        h = torch.arange((-self.scale + 1) / 2, (self.scale - 1) / 2 + 1) / self.scale
        return torch.stack(torch.meshgrid([h, h], indexing='ij')).transpose(1, 2).repeat(1, self.groups, 1).reshape(1, -1, 1, 1)

    def forward(self, x):
        offset = self.offset(x) * self.scope(x).sigmoid() * 0.5 + self.init_pos
        B, _, H, W = offset.shape
        offset = offset.view(B, 2, -1, H, W)
        coords_h = torch.arange(H) + 0.5
        coords_w = torch.arange(W) + 0.5

        coords = (
            torch.stack(torch.meshgrid([coords_w, coords_h], indexing='ij'))
            .transpose(1, 2)
            .unsqueeze(1)
            .unsqueeze(0)
            .type(x.dtype)
            .to(x.device, non_blocking=True)
        )
        normalizer = torch.tensor([W, H], dtype=x.dtype, device=x.device, pin_memory=True).view(1, 2, 1, 1, 1)
        coords = 2 * (coords + offset) / normalizer - 1

        coords = (
            F.pixel_shuffle(coords.reshape(B, -1, H, W), self.scale)
            .view(B, 2, -1, self.scale * H, self.scale * W)
            .permute(0, 2, 3, 4, 1)
            .contiguous()
            .flatten(0, 1)
        )
        output = F.grid_sample(
            x.reshape(B * self.groups, -1, H, W),
            coords,
            mode='bilinear',
            align_corners=False,
            padding_mode='border',
        ).view(B, -1, self.scale * H, self.scale * W)

        if self.end_convolution:
            output = self.end_conv(output)

        return output



import math
from typing import Literal

import torch
import torch.nn.functional as F
from torch import Tensor, nn


SampleMods = Literal['conv', 'pixelshuffledirect', 'pixelshuffle', 'nearest+conv', 'dysample']


class DySample(nn.Module):
    """Adapted from 'Learning to Upsample by Learning to Sample':
    https://arxiv.org/abs/2308.15085
    https://github.com/tiny-smart/dysample
    """

    def __init__(
        self,
        in_channels: int,
        out_ch: int,
        scale: int = 2,
        groups: int = 4,
        end_convolution: bool = True,
    ) -> None:
        super().__init__()

        if in_channels <= groups or in_channels % groups != 0:
            msg = 'Incorrect in_channels and groups values.'
            raise ValueError(msg)

        out_channels = 2 * groups * scale**2
        self.scale = scale
        self.groups = groups
        self.end_convolution = end_convolution
        if end_convolution:
            self.end_conv = nn.Conv2d(in_channels, out_ch, kernel_size=1)

        self.offset = nn.Conv2d(in_channels, out_channels, 1)
        self.scope = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        if self.training:
            nn.init.trunc_normal_(self.offset.weight, std=0.02)
            nn.init.constant_(self.scope.weight, val=0)

        self.register_buffer('init_pos', self._init_pos())

    def _init_pos(self) -> Tensor:
        h = torch.arange((-self.scale + 1) / 2, (self.scale - 1) / 2 + 1) / self.scale
        return torch.stack(torch.meshgrid([h, h], indexing='ij')).transpose(1, 2).repeat(1, self.groups, 1).reshape(1, -1, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        offset = self.offset(x) * self.scope(x).sigmoid() * 0.5 + self.init_pos
        B, _, H, W = offset.shape
        offset = offset.view(B, 2, -1, H, W)
        coords_h = torch.arange(H) + 0.5
        coords_w = torch.arange(W) + 0.5

        coords = (
            torch.stack(torch.meshgrid([coords_w, coords_h], indexing='ij'))
            .transpose(1, 2)
            .unsqueeze(1)
            .unsqueeze(0)
            .type(x.dtype)
            .to(x.device, non_blocking=True)
        )
        normalizer = torch.tensor([W, H], dtype=x.dtype, device=x.device, pin_memory=True).view(1, 2, 1, 1, 1)
        coords = 2 * (coords + offset) / normalizer - 1

        coords = (
            F.pixel_shuffle(coords.reshape(B, -1, H, W), self.scale)
            .view(B, 2, -1, self.scale * H, self.scale * W)
            .permute(0, 2, 3, 4, 1)
            .contiguous()
            .flatten(0, 1)
        )
        output = F.grid_sample(
            x.reshape(B * self.groups, -1, H, W),
            coords,
            mode='bilinear',
            align_corners=False,
            padding_mode='border',
        ).view(B, -1, self.scale * H, self.scale * W)

        if self.end_convolution:
            output = self.end_conv(output)

        return output


class UniUpsample(nn.Sequential):
    def __init__(
        self,
        upsample: SampleMods,
        scale: int = 2,
        in_dim: int = 64,
        out_dim: int = 3,
        mid_dim: int = 64,  # Only pixelshuffle
        group: int = 4,  # Only DySample
    ) -> None:
        m = []

        if scale == 1 or upsample == 'conv':
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'pixelshuffledirect':
            m.extend([nn.Conv2d(in_dim, out_dim * scale**2, 3, 1, 1), nn.PixelShuffle(scale)])
        elif upsample == 'pixelshuffle':
            m.extend([nn.Conv2d(in_dim, mid_dim, 3, 1, 1), nn.LeakyReLU(inplace=True)])
            if (scale & (scale - 1)) == 0:  # scale = 2^n
                for _ in range(int(math.log2(scale))):
                    m.extend([nn.Conv2d(mid_dim, 4 * mid_dim, 3, 1, 1), nn.PixelShuffle(2)])
            elif scale == 3:
                m.extend([nn.Conv2d(mid_dim, 9 * mid_dim, 3, 1, 1), nn.PixelShuffle(3)])
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(mid_dim, out_dim, 3, 1, 1))
        elif upsample == 'nearest+conv':
            if (scale & (scale - 1)) == 0:
                for _ in range(int(math.log2(scale))):
                    m.extend(
                        (
                            nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                            nn.Upsample(scale_factor=2),
                            nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        )
                    )
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            elif scale == 3:
                m.extend(
                    (
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.Upsample(scale_factor=scale),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                        nn.Conv2d(in_dim, in_dim, 3, 1, 1),
                        nn.LeakyReLU(negative_slope=0.2, inplace=True),
                    )
                )
            else:
                raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
            m.append(nn.Conv2d(in_dim, out_dim, 3, 1, 1))
        elif upsample == 'dysample':
            m.append(DySample(in_dim, out_dim, scale, group))
        else:
            raise ValueError(f'An invalid Upsample was selected. Please choose one of {SampleMods}')
        super().__init__(*m)

        self.register_buffer(
            'MetaUpsample',
            torch.tensor(
                [
                    1,  # Block version, if you change something, please number from the end so that you can distinguish between authorized changes and third parties
                    list(SampleMods.__args__).index(upsample),  # UpSample method index
                    scale,
                    in_dim,
                    out_dim,
                    mid_dim,
                    group,
                ],
                dtype=torch.uint8,
            ),
        )


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6) -> None:
        super().__init__()
        self.eps = eps
        self.scale = nn.Parameter(torch.ones(dim))
        self.offset = nn.Parameter(torch.zeros(dim))

    def forward(self, x):
        norm_x = x.norm(2, dim=1, keepdim=True)
        d_x = x.size(1)
        rms_x = norm_x * (d_x ** (-1.0 / 2))
        x_normed = x / (rms_x + self.eps)
        return self.scale[..., None, None] * x_normed + self.offset[..., None, None]


class InceptionDWConv2d(nn.Module):
    """Inception depthweise convolution"""

    def __init__(self, in_channels, square_kernel_size=3, band_kernel_size=11, branch_ratio=0.125) -> None:
        super().__init__()

        gc = int(in_channels * branch_ratio)  # channel numbers of a convolution branch
        self.dwconv_hw = nn.Conv2d(gc, gc, square_kernel_size, padding=square_kernel_size // 2, groups=gc)
        self.dwconv_w = nn.Conv2d(
            gc,
            gc,
            kernel_size=(1, band_kernel_size),
            padding=(0, band_kernel_size // 2),
            groups=gc,
        )
        self.dwconv_h = nn.Conv2d(
            gc,
            gc,
            kernel_size=(band_kernel_size, 1),
            padding=(band_kernel_size // 2, 0),
            groups=gc,
        )
        self.split_indexes = (in_channels - 3 * gc, gc, gc, gc)

    def forward(self, x):
        x_id, x_hw, x_w, x_h = torch.split(x, self.split_indexes, dim=1)
        return torch.cat(
            (x_id, self.dwconv_hw(x_hw), self.dwconv_w(x_w), self.dwconv_h(x_h)),
            dim=1,
        )


def l2_norm(x):
    return torch.einsum('bcn, bn->bcn', x, 1 / torch.norm(x, p=2, dim=-2))


class Attention(nn.Module):
    def __init__(self, in_places, scale=8, eps=1e-6) -> None:
        super().__init__()
        # self.gamma = nn.Parameter(torch.zeros(1,in_places,1,1))
        self.in_places = in_places
        self.l2_norm = l2_norm
        self.eps = eps

        self.query_conv = nn.Conv2d(in_channels=in_places, out_channels=in_places // scale, kernel_size=1)
        self.key_conv = nn.Conv2d(in_channels=in_places, out_channels=in_places // scale, kernel_size=1)
        self.value_conv = nn.Conv2d(in_channels=in_places, out_channels=in_places, kernel_size=1)

    def forward(self, x):
        # Apply the feature map to the queries and keys
        batch_size, chnnels, height, width = x.shape
        Q = self.query_conv(x).view(batch_size, -1, width * height)
        K = self.key_conv(x).view(batch_size, -1, width * height)
        V = self.value_conv(x).view(batch_size, -1, width * height)

        Q = self.l2_norm(Q).permute(-3, -1, -2)
        K = self.l2_norm(K)

        tailor_sum = 1 / (width * height + torch.einsum('bnc, bc->bn', Q, torch.sum(K, dim=-1) + self.eps))
        value_sum = torch.einsum('bcn->bc', V).unsqueeze(-1)
        value_sum = value_sum.expand(-1, chnnels, width * height)

        matrix = torch.einsum('bmn, bcn->bmc', K, V)
        matrix_sum = value_sum + torch.einsum('bnm, bmc->bcn', Q, matrix)

        weight_value = torch.einsum('bcn, bn->bcn', matrix_sum, tailor_sum)
        weight_value = weight_value.view(batch_size, chnnels, height, width)
        return weight_value.contiguous()


class GatedCNNBlock(nn.Module):
    r"""
    modernized mambaout main unit
    https://github.com/yuweihao/MambaOut/blob/main/models/mambaout.py#L119
    """

    def __init__(
        self,
        dim: int = 64,
        expansion_ratio: float = 1.5,
        conv_ratio: float = 1,
        att=False,
    ) -> None:
        super().__init__()
        self.norm = RMSNorm(dim)
        hidden = int(expansion_ratio * dim)
        self.fc1 = nn.Conv2d(dim, hidden * 2, 1, 1)

        self.act = nn.Mish()
        conv_channels = int(conv_ratio * dim)
        self.split_indices = [hidden, hidden - conv_channels, conv_channels]
        self.token_mix = Attention(conv_channels, 16) if att else InceptionDWConv2d(dim)
        self.fc2 = nn.Conv2d(hidden, dim, 1, 1, 0)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.norm(x)
        g, i, c = torch.split(self.fc1(x), self.split_indices, dim=1)
        c = self.token_mix(c)
        x = self.act(g) * torch.cat((i, c), dim=1)
        x = self.act(self.fc2(x))
        return x


class SimpleGate(nn.Module):
    @staticmethod
    def forward(x):
        x1, x2 = x.chunk(2, dim=1)
        return x1 * x2


class MetaGated(nn.Module):
    def __init__(self, dim) -> None:
        super().__init__()
        hidden = dim * 2
        self.local = nn.Sequential(
            RMSNorm(dim),
            nn.Conv2d(dim, hidden, 1),
            nn.Conv2d(hidden, hidden, 3, 1, 1, groups=dim),
            SimpleGate(),
        )
        self.sca = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels=dim, out_channels=dim, kernel_size=1),
        )
        self.glob = GatedCNNBlock(dim)
        self.gamma0 = nn.Parameter(torch.ones([1, dim, 1, 1]), requires_grad=True)
        self.gamma1 = nn.Parameter(torch.ones([1, dim, 1, 1]), requires_grad=True)
        self.apply(self._init_weights)

    @staticmethod
    def _init_weights(m) -> None:
        if isinstance(m, nn.Conv2d | nn.Linear):
            nn.init.trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        short = x
        x = self.local(x)
        x = x * self.sca(x)
        x = x * self.gamma0 + short
        del short
        x = self.glob(x) * self.gamma1 + x
        return x


class Down(nn.Sequential):
    def __init__(self, dim) -> None:
        super().__init__(nn.Conv2d(dim, dim // 2, 3, 1, 1, bias=False), nn.PixelUnshuffle(2))


class Upsample(nn.Sequential):
    def __init__(self, dim) -> None:
        super().__init__(nn.Conv2d(dim, dim * 2, 3, 1, 1, bias=False), nn.PixelShuffle(2))


class Block(nn.Module):
    def __init__(self, dim, num_gated, down=True) -> None:
        super().__init__()

        if down:
            self.gated = nn.Sequential(*[MetaGated(dim) for _ in range(num_gated)])
            self.scale = Down(dim)

        else:
            self.scale = Upsample(dim)
            self.gated = nn.Sequential(*[MetaGated(dim // 2) for _ in range(num_gated)])
            self.shor = nn.Conv2d(int(dim), dim // 2, 1, 1, 0)
        self.down = down

    def forward(self, x, short=None):
        if self.down:
            x = self.gated(x)
            return self.scale(x), x
        else:
            x = torch.cat([self.scale(x), short], dim=1)
            x = self.shor(x)
            return self.gated(x)


class ConvBlock(nn.Module):
    r"""https://github.com/joshyZhou/AST/blob/main/model.py#L22"""

    def __init__(self, in_channel: int, out_channel: int, strides: int = 1) -> None:
        super().__init__()
        self.strides = strides
        self.in_channel = in_channel
        self.out_channel = out_channel
        self.block = nn.Sequential(
            nn.Conv2d(in_channel, out_channel, kernel_size=3, stride=strides, padding=1),
            nn.Mish(True),
            nn.Conv2d(out_channel, out_channel, kernel_size=3, stride=strides, padding=1),
            nn.Mish(True),
        )
        self.conv11 = nn.Conv2d(in_channel, out_channel, kernel_size=1, stride=strides, padding=0)

    def forward(self, x):
        out1 = self.block(x)
        out2 = self.conv11(x)
        out = out1 + out2
        return out


class GateRV2(nn.Module):
    def __init__(
        self,
        in_ch=3,
        dim=48,
        enc_blocks=(2, 2, 4, 6),
        dec_blocks=(2, 2, 2, 2),
        num_latent=10,
        scale=1,
        upsample: SampleMods = 'pixelshuffledirect',
        upsample_mid_dim=32,
        **kwargs,
    ) -> None:
        super().__init__()
        self.scale = 1
        self.in_to_dim = nn.Conv2d(in_ch, dim, 3, 1, 1)

        self.encode = nn.ModuleList([Block(dim * (2**i), enc_blocks[i]) for i in range(len(enc_blocks))])
        self.latent = nn.Sequential(
            *[
                GatedCNNBlock(
                    dim * (2 ** len(enc_blocks)),
                    expansion_ratio=1.5,
                    conv_ratio=1.00,
                    att=True,
                )
                for _ in range(num_latent)
            ]
        )
        self.decode = nn.ModuleList(
            [
                Block(
                    dim * (2 ** (len(dec_blocks) - i)),
                    dec_blocks[i],
                    False,
                )
                for i in range(len(dec_blocks))
            ]
        )
        self.pad = 2 ** (len(enc_blocks))

        if scale != 1:
            self.dim_to_in = nn.Identity()
            self.short_to_dim = ConvBlock(in_ch, dim)
            self.upsample = UniUpsample(upsample, scale, dim, in_ch, upsample_mid_dim)
        else:
            self.dim_to_in = nn.Conv2d(dim, in_ch, 3, 1, 1)
            self.short_to_dim = nn.Identity()
            self.upsample = nn.Identity()

    def load_state_dict(self, state_dict, *args, **kwargs):
        if 'upsample.MetaUpsample' in state_dict:
            state_dict['upsample.MetaUpsample'] = self.upsample.MetaUpsample
        return super().load_state_dict(state_dict, *args, **kwargs)

    def check_img_size(self, x, resolution: tuple[int, int]):
        scaled_size = self.pad
        mod_pad_h = (scaled_size - resolution[0] % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - resolution[1] % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, inp):
        B, C, H, W = inp.shape
        inp = self.check_img_size(inp, (H, W))
        x = self.in_to_dim(inp)
        shorts = []
        for block in self.encode:
            x, short = block(x)
            shorts.append(short)
        x = self.latent(x)
        len_block = len(self.decode)
        shorts.reverse()
        for index in range(len_block):
            x = self.decode[index](x, shorts[index])
        x = self.dim_to_in(x) + self.short_to_dim(inp)

        return self.upsample(x)[:, :, : H * self.scale, : W * self.scale]



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from ...utilities import to_2tuple


class ChannelAttention(nn.Module):
    """Channel attention used in RCAN.
    Args:
        num_feat (int): Channel number of intermediate features.
        squeeze_factor (int): Channel squeeze factor. Default: 16.
    """

    def __init__(self, num_feat, squeeze_factor=16):
        super().__init__()
        self.attention = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(num_feat, num_feat // squeeze_factor, 1, padding=0),
            nn.ReLU(inplace=True),
            nn.Conv2d(num_feat // squeeze_factor, num_feat, 1, padding=0),
            nn.Sigmoid(),
        )

    def forward(self, x):
        y = self.attention(x)
        return x * y


class CAB(nn.Module):
    def __init__(self, num_feat, compress_ratio=3, squeeze_factor=30):
        super().__init__()

        self.cab = nn.Sequential(
            nn.Conv2d(num_feat, num_feat // compress_ratio, 3, 1, 1),
            nn.GELU(),
            nn.Conv2d(num_feat // compress_ratio, num_feat, 3, 1, 1),
            ChannelAttention(num_feat, squeeze_factor),
        )

    def forward(self, x):
        return self.cab(x)


class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


def window_partition(x, window_size):
    """
    Args:
        x: (B, H, W, C)
        window_size (int): window size
    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    B, H, W, C = x.shape
    x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
    return windows


def window_reverse(windows, window_size, H, W):
    """
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size
        H (int): Height of image
        W (int): Width of image
    Returns:
        x: (B, H, W, C)
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(B, H // window_size, W // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x


class WindowAttention(nn.Module):
    r"""Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports both of shifted and non-shifted window.
    Args:
        dim (int): Number of input channels.
        window_size (tuple[int]): The height and width of the window.
        num_heads (int): Number of attention heads.
        qkv_bias (bool, optional):  If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set
        attn_drop (float, optional): Dropout ratio of attention weight. Default: 0.0
        proj_drop (float, optional): Dropout ratio of output. Default: 0.0
    """

    def __init__(
        self,
        dim,
        window_size,
        num_heads,
        qkv_bias=True,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        # define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(  # type: ignore
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1), num_heads)
        )  # 2*Wh-1 * 2*Ww-1, nH

        # get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w], indexing='ij'))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += self.window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
        self.register_buffer('relative_position_index', relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)

        self.proj_drop = nn.Dropout(proj_drop)

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask=None):
        """
        Args:
            x: input features with shape of (num_windows*B, N, C)
            mask: (0/-inf) mask with shape of (num_windows, Wh*Ww, Wh*Ww) or None
        """
        B_, N, C = x.shape
        qkv = self.qkv(x).reshape(B_, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = (
            qkv[0],
            qkv[1],
            qkv[2],
        )  # make torchscript happy (cannot use tensor as tuple)

        q = q * self.scale
        attn = q @ k.transpose(-2, -1)

        relative_position_bias = self.relative_position_bias_table[
            self.relative_position_index.view(-1)  # type: ignore
        ].view(
            self.window_size[0] * self.window_size[1],
            self.window_size[0] * self.window_size[1],
            -1,
        )  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.view(B_ // nW, nW, self.num_heads, N, N) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    def extra_repr(self) -> str:
        return f'dim={self.dim}, window_size={self.window_size}, num_heads={self.num_heads}'


class RDG(nn.Module):
    def __init__(
        self,
        dim,
        input_resolution,
        num_heads,
        window_size,
        mlp_ratio,
        qkv_bias,
        qk_scale,
        drop,
        attn_drop,
        drop_path,
        norm_layer,
        gc,
        patch_size,
        img_size,
    ):
        super().__init__()

        self.swin1 = SwinTransformerBlock(
            dim=dim,
            input_resolution=input_resolution,
            num_heads=num_heads,
            window_size=window_size,
            shift_size=0,  # For first block
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            drop=drop,
            attn_drop=attn_drop,
            drop_path=drop_path[0] if isinstance(drop_path, list) else drop_path,
            norm_layer=norm_layer,  # type: ignore
        )
        self.adjust1 = nn.Conv2d(dim, gc, 1)

        self.swin2 = SwinTransformerBlock(
            dim + gc,
            input_resolution=input_resolution,
            num_heads=num_heads - ((dim + gc) % num_heads),
            window_size=window_size,
            shift_size=window_size // 2,  # For first block
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            drop=drop,
            attn_drop=attn_drop,
            drop_path=drop_path[0] if isinstance(drop_path, list) else drop_path,
            norm_layer=norm_layer,  # type: ignore
        )
        self.adjust2 = nn.Conv2d(dim + gc, gc, 1)

        self.swin3 = SwinTransformerBlock(
            dim + 2 * gc,
            input_resolution=input_resolution,
            num_heads=num_heads - ((dim + 2 * gc) % num_heads),
            window_size=window_size,
            shift_size=0,  # For first block
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            drop=drop,
            attn_drop=attn_drop,
            drop_path=drop_path[0] if isinstance(drop_path, list) else drop_path,
            norm_layer=norm_layer,  # type: ignore
        )
        self.adjust3 = nn.Conv2d(dim + gc * 2, gc, 1)

        self.swin4 = SwinTransformerBlock(
            dim + 3 * gc,
            input_resolution=input_resolution,
            num_heads=num_heads - ((dim + 3 * gc) % num_heads),
            window_size=window_size,
            shift_size=window_size // 2,  # For first block
            mlp_ratio=1,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            drop=drop,
            attn_drop=attn_drop,
            drop_path=drop_path[0] if isinstance(drop_path, list) else drop_path,
            norm_layer=norm_layer,  # type: ignore
        )
        self.adjust4 = nn.Conv2d(dim + gc * 3, gc, 1)

        self.swin5 = SwinTransformerBlock(
            dim + 4 * gc,
            input_resolution=input_resolution,
            num_heads=num_heads - ((dim + 4 * gc) % num_heads),
            window_size=window_size,
            shift_size=0,  # For first block
            mlp_ratio=1,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            drop=drop,
            attn_drop=attn_drop,
            drop_path=drop_path[0] if isinstance(drop_path, list) else drop_path,
            norm_layer=norm_layer,  # type: ignore
        )
        self.adjust5 = nn.Conv2d(dim + gc * 4, dim, 1)

        self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

        self.pe = PatchEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=0,
            embed_dim=dim,
            norm_layer=None,
        )

        self.pue = PatchUnEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=0,
            embed_dim=dim,
            norm_layer=None,
        )

    def forward(self, x, xsize):
        x1 = self.pe(self.lrelu(self.adjust1(self.pue(self.swin1(x, xsize), xsize))))
        x2 = self.pe(self.lrelu(self.adjust2(self.pue(self.swin2(torch.cat((x, x1), -1), xsize), xsize))))
        x3 = self.pe(self.lrelu(self.adjust3(self.pue(self.swin3(torch.cat((x, x1, x2), -1), xsize), xsize))))
        x4 = self.pe(self.lrelu(self.adjust4(self.pue(self.swin4(torch.cat((x, x1, x2, x3), -1), xsize), xsize))))
        x5 = self.pe(self.adjust5(self.pue(self.swin5(torch.cat((x, x1, x2, x3, x4), -1), xsize), xsize)))

        return x5 * 0.2 + x


class SwinTransformerBlock(nn.Module):
    r"""Swin Transformer Block.
    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resulotion.
        num_heads (int): Number of attention heads.
        window_size (int): Window size.
        shift_size (int): Shift size for SW-MSA.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float, optional): Stochastic depth rate. Default: 0.0
        act_layer (nn.Module, optional): Activation layer. Default: nn.GELU
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(
        self,
        dim,
        input_resolution,
        num_heads,
        window_size=7,
        shift_size=0,
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        if min(self.input_resolution) <= self.window_size:
            # if window size is larger than input resolution, we don't partition windows
            self.shift_size = 0
            self.window_size = min(self.input_resolution)
        assert 0 <= self.shift_size < self.window_size, 'shift_size must in 0-window_size'

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim,
            window_size=to_2tuple(self.window_size),
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
        )

        self.drop_path = nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

        if self.shift_size > 0:
            attn_mask = self.calculate_mask(self.input_resolution)
        else:
            attn_mask = None

        self.register_buffer('attn_mask', attn_mask)

    def calculate_mask(self, x_size):
        # calculate attention mask for SW-MSA
        H, W = x_size
        img_mask = torch.zeros((1, H, W, 1))  # 1 H W 1
        h_slices = (
            slice(0, -self.window_size),
            slice(-self.window_size, -self.shift_size),
            slice(-self.shift_size, None),
        )
        w_slices = (
            slice(0, -self.window_size),
            slice(-self.window_size, -self.shift_size),
            slice(-self.shift_size, None),
        )
        cnt = 0
        for h in h_slices:
            for w in w_slices:
                img_mask[:, h, w, :] = cnt
                cnt += 1

        mask_windows = window_partition(img_mask, self.window_size)  # nW, window_size, window_size, 1
        mask_windows = mask_windows.view(-1, self.window_size * self.window_size)
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0, -100.0).masked_fill(attn_mask == 0, 0.0)

        return attn_mask

    def forward(self, x, x_size):
        H, W = x_size
        B, _L, C = x.shape
        # assert L == H * W, "input feature has wrong size"

        shortcut = x
        x = self.norm1(x)
        x = x.view(B, H, W, C)

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
        else:
            shifted_x = x

        # partition windows
        x_windows = window_partition(shifted_x, self.window_size)  # nW*B, window_size, window_size, C
        x_windows = x_windows.view(-1, self.window_size * self.window_size, C)  # nW*B, window_size*window_size, C

        # W-MSA/SW-MSA (to be compatible for testing on images whose shapes are the multiple of window size
        if self.input_resolution == x_size:
            attn_windows = self.attn(x_windows, mask=self.attn_mask)  # nW*B, window_size*window_size, C
        else:
            attn_windows = self.attn(x_windows, mask=self.calculate_mask(x_size).to(x.device))

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size, C)
        shifted_x = window_reverse(attn_windows, self.window_size, H, W)  # B H' W' C

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2))
        else:
            x = shifted_x
        x = x.view(B, H * W, C)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x

    def extra_repr(self) -> str:
        return (
            f'dim={self.dim}, input_resolution={self.input_resolution}, num_heads={self.num_heads}, '
            f'window_size={self.window_size}, shift_size={self.shift_size}, mlp_ratio={self.mlp_ratio}'
        )


class PatchMerging(nn.Module):
    r"""Patch Merging Layer.

    Args:
        input_resolution (tuple[int]): Resolution of input feature.
        dim (int): Number of input channels.
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, input_resolution, dim: int, norm_layer=nn.LayerNorm):
        super().__init__()
        self.input_resolution = input_resolution
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x):
        """
        x: b, h*w, c
        """
        h, w = self.input_resolution
        b, seq_len, c = x.shape
        assert seq_len == h * w, 'input feature has wrong size'
        assert h % 2 == 0 and w % 2 == 0, f'x size ({h}*{w}) are not even.'

        x = x.view(b, h, w, c)

        x0 = x[:, 0::2, 0::2, :]  # b h/2 w/2 c
        x1 = x[:, 1::2, 0::2, :]  # b h/2 w/2 c
        x2 = x[:, 0::2, 1::2, :]  # b h/2 w/2 c
        x3 = x[:, 1::2, 1::2, :]  # b h/2 w/2 c
        x = torch.cat([x0, x1, x2, x3], -1)  # b h/2 w/2 4*c
        x = x.view(b, -1, 4 * c)  # b h/2*w/2 4*c

        x = self.norm(x)
        x = self.reduction(x)

        return x


class PatchEmbed(nn.Module):
    r"""Image to Patch Embedding

    Args:
        img_size (int): Image size.  Default: 224.
        patch_size (int): Patch token size. Default: 4.
        in_chans (int): Number of input image channels. Default: 3.
        embed_dim (int): Number of linear projection output channels. Default: 96.
        norm_layer (nn.Module, optional): Normalization layer. Default: None
    """

    def __init__(self, img_size=224, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        patches_resolution = [
            img_size[0] // patch_size[0],
            img_size[1] // patch_size[1],
        ]
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x):
        x = x.flatten(2).transpose(1, 2)  # structured as [B, num_patches, C]
        if self.norm is not None:
            x = self.norm(x)  # 归一化
        return x


class PatchUnEmbed(nn.Module):
    r"""Image to Patch Unembedding

    args:
        img_size (int): image size, default 224*224.
        patch_size (int): Patch token sizd, default 4*4.
        in_chans (int): num image channels, default 3.
        embed_dim (int): num channels for linear projection output, default 96
        norm_layer (nn.Module, optional): normalization layer, default None.
    """

    def __init__(self, img_size=224, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None):
        super().__init__()
        img_size = to_2tuple(img_size)  # 图像的大小，默认为 224*224
        patch_size = to_2tuple(patch_size)  # Patch token 的大小，默认为 4*4
        patches_resolution = [
            img_size[0] // patch_size[0],
            img_size[1] // patch_size[1],
        ]  # patch 的分辨率
        self.img_size = img_size
        self.patch_size = patch_size
        self.patches_resolution = patches_resolution
        self.num_patches = patches_resolution[0] * patches_resolution[1]  # patch 的个数，num_patches

        self.in_chans = in_chans  # 输入图像的通道数
        self.embed_dim = embed_dim  # 线性 projection 输出的通道数

    def forward(self, x, x_size):
        B, _HW, _C = x.shape  # 输入 x 的结构
        x = x.transpose(1, 2).view(B, -1, x_size[0], x_size[1])  # 输出结构为 [B, Ph*Pw, C]
        return x


class Upsample(nn.Sequential):
    """Upsample module.

    Args:
        scale (int): Scale factor. Supported scales: 2^n and 3.
        num_feat (int): Channel number of intermediate features.
    """

    def __init__(self, scale, num_feat):
        m = []
        if (scale & (scale - 1)) == 0:  # scale = 2^n
            for _ in range(int(math.log2(scale))):
                m.append(nn.Conv2d(num_feat, 4 * num_feat, 3, 1, 1))
                m.append(nn.PixelShuffle(2))
        elif scale == 3:
            m.append(nn.Conv2d(num_feat, 9 * num_feat, 3, 1, 1))
            m.append(nn.PixelShuffle(3))
        else:
            raise ValueError(f'scale {scale} is not supported. Supported scales: 2^n and 3.')
        super().__init__(*m)


class DRCT(nn.Module):
    hyperparameters = {}

    def __init__(
        self,
        *,
        img_size=64,
        patch_size=1,
        in_chans=3,
        embed_dim=180,
        depths=(6, 6, 6, 6, 6, 6),
        num_heads=(6, 6, 6, 6, 6, 6),
        window_size=16,
        mlp_ratio=2.0,
        qkv_bias=True,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.1,
        norm_layer=nn.LayerNorm,
        ape=False,
        patch_norm=True,
        upscale=1,
        img_range=1.0,
        upsampler='',
        resi_connection='1conv',
        gc=32,
    ):
        super().__init__()

        self.window_size = window_size
        self.shift_size = window_size // 2

        num_in_ch = in_chans
        num_out_ch = in_chans
        num_feat = 64
        self.img_range = img_range
        if in_chans == 3:
            rgb_mean = (0.4488, 0.4371, 0.4040)
            self.mean = torch.Tensor(rgb_mean).view(1, 3, 1, 1)
        else:
            self.mean = torch.zeros(1, 1, 1, 1)
        self.upscale = upscale
        self.upsampler = upsampler

        # ------------------------- 1, shallow feature extraction ------------------------- #
        self.conv_first = nn.Conv2d(num_in_ch, embed_dim, 3, 1, 1)

        # ------------------------- 2, deep feature extraction ------------------------- #
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.ape = ape
        self.patch_norm = patch_norm
        self.num_features = embed_dim
        self.mlp_ratio = mlp_ratio

        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=embed_dim,
            embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None,
        )
        num_patches = self.patch_embed.num_patches
        patches_resolution = self.patch_embed.patches_resolution
        self.patches_resolution = patches_resolution

        # merge non-overlapping patches into image
        self.patch_unembed = PatchUnEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=embed_dim,
            embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None,
        )

        # absolute position embedding
        if self.ape:
            self.absolute_pos_embed = nn.Parameter(  # type: ignore
                torch.zeros(1, num_patches, embed_dim)
            )

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

        # build
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers):
            layer = RDG(
                dim=embed_dim,
                input_resolution=(patches_resolution[0], patches_resolution[1]),
                num_heads=num_heads[i_layer],
                window_size=window_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[sum(depths[:i_layer]) : sum(depths[: i_layer + 1])],
                norm_layer=norm_layer,
                gc=gc,
                img_size=img_size,
                patch_size=patch_size,
            )

            self.layers.append(layer)
        self.norm = norm_layer(self.num_features)

        # build the last conv layer in deep feature extraction
        if resi_connection == '1conv':
            self.conv_after_body = nn.Conv2d(embed_dim, embed_dim, 3, 1, 1)
        elif resi_connection == 'identity':
            self.conv_after_body = nn.Identity()

        # ------------------------- 3, high quality image reconstruction ------------------------- #
        if self.upsampler == 'pixelshuffle':
            # for classical SR
            self.conv_before_upsample = nn.Sequential(nn.Conv2d(embed_dim, num_feat, 3, 1, 1), nn.LeakyReLU(inplace=True))
            self.upsample = Upsample(upscale, num_feat)
            self.conv_last = nn.Conv2d(num_feat, num_out_ch, 3, 1, 1)

    @torch.jit.ignore  # type: ignore
    def no_weight_decay(self):
        return {'absolute_pos_embed'}

    @torch.jit.ignore  # type: ignore
    def no_weight_decay_keywords(self):
        return {'relative_position_bias_table'}

    def forward_features(self, x):
        x_size = (x.shape[2], x.shape[3])

        x = self.patch_embed(x)
        if self.ape:
            x = x + self.absolute_pos_embed
        x = self.pos_drop(x)

        for layer in self.layers:
            x = layer(x, x_size)

        x = self.norm(x)  # b seq_len c
        x = self.patch_unembed(x, x_size)

        return x

    def check_img_size(self, x, resolution):
        h, w = resolution
        scaled_size = self.window_size
        mod_pad_h = (scaled_size - h % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - w % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')

    def forward(self, x):
        self.mean = self.mean.type_as(x)
        x = (x - self.mean) * self.img_range
        _, _, h, w = x.shape
        x = self.check_img_size(x, (h, w))
        if self.upsampler == 'pixelshuffle':
            # for classical SR
            x = self.conv_first(x)
            x = self.conv_after_body(self.forward_features(x)) + x
            x = self.conv_before_upsample(x)
            x = self.conv_last(self.upsample(x))

        x = x / self.img_range + self.mean

        return x[:, :, : h * self.upscale, : w * self.upscale]



from __future__ import annotations

from collections import OrderedDict
from typing import Literal

import torch
import torch.nn.functional as F
from torch import nn as nn


def _make_pair(value):
    if isinstance(value, int):
        return (value, value)
    return value


def conv_layer(in_channels, out_channels, kernel_size, bias=True):
    """
    Re-write convolution layer for adaptive `padding`.
    """
    kernel_size = _make_pair(kernel_size)
    padding = (int((kernel_size[0] - 1) / 2), int((kernel_size[1] - 1) / 2))
    return nn.Conv2d(in_channels, out_channels, kernel_size, padding=padding, bias=bias)


def sequential(*args):
    """
    Modules will be added to the a Sequential Container in the order they
    are passed.

    Parameters
    ----------
    args: Definition of Modules in order.
    -------
    """
    if len(args) == 1:
        if isinstance(args[0], OrderedDict):
            raise NotImplementedError('sequential does not support OrderedDict input.')
        return args[0]
    modules = []
    for module in args:
        if isinstance(module, nn.Sequential):
            for submodule in module.children():
                modules.append(submodule)
        elif isinstance(module, nn.Module):
            modules.append(module)
    return nn.Sequential(*modules)


def pixelshuffle_block(in_channels, out_channels, upscale_factor=2, kernel_size=3):
    """
    Upsample features according to `upscale_factor`.
    """
    conv = conv_layer(in_channels, out_channels * (upscale_factor**2), kernel_size)
    pixel_shuffle = nn.PixelShuffle(upscale_factor)
    return sequential(conv, pixel_shuffle)


class Conv3XC(nn.Module):
    def __init__(
        self,
        c_in: int,
        c_out: int,
        gain1=1,
        s=1,
        bias: Literal[True] = True,
        relu=False,
    ):
        super().__init__()
        self.weight_concat = None
        self.bias_concat = None
        self.update_params_flag = False
        self.stride = s
        self.has_relu = relu
        gain = gain1

        self.sk = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=1,
            padding=0,
            stride=s,
            bias=bias,
        )
        self.conv = nn.Sequential(
            nn.Conv2d(
                in_channels=c_in,
                out_channels=c_in * gain,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_in * gain,
                out_channels=c_out * gain,
                kernel_size=3,
                stride=s,
                padding=0,
                bias=bias,
            ),
            nn.Conv2d(
                in_channels=c_out * gain,
                out_channels=c_out,
                kernel_size=1,
                padding=0,
                bias=bias,
            ),
        )

        self.eval_conv = nn.Conv2d(
            in_channels=c_in,
            out_channels=c_out,
            kernel_size=3,
            padding=1,
            stride=s,
            bias=bias,
        )

        if not self.training:
            self.eval_conv.weight.requires_grad = False
            self.eval_conv.bias.requires_grad = False  # type: ignore
            self.update_params()

    def update_params(self):
        w1 = self.conv[0].weight.data.clone().detach()
        b1 = self.conv[0].bias.data.clone().detach()
        w2 = self.conv[1].weight.data.clone().detach()
        b2 = self.conv[1].bias.data.clone().detach()
        w3 = self.conv[2].weight.data.clone().detach()
        b3 = self.conv[2].bias.data.clone().detach()

        w = F.conv2d(w1.flip(2, 3).permute(1, 0, 2, 3), w2, padding=2, stride=1).flip(2, 3).permute(1, 0, 2, 3)
        b = (w2 * b1.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b2

        self.weight_concat = F.conv2d(w.flip(2, 3).permute(1, 0, 2, 3), w3, padding=0, stride=1).flip(2, 3).permute(1, 0, 2, 3)
        self.bias_concat = (w3 * b.reshape(1, -1, 1, 1)).sum((1, 2, 3)) + b3

        sk_w = self.sk.weight.data.clone().detach()
        sk_b = self.sk.bias.data.clone().detach()  # type: ignore
        target_kernel_size = 3

        H_pixels_to_pad = (target_kernel_size - 1) // 2
        W_pixels_to_pad = (target_kernel_size - 1) // 2
        sk_w = F.pad(sk_w, [H_pixels_to_pad, H_pixels_to_pad, W_pixels_to_pad, W_pixels_to_pad])

        self.weight_concat = self.weight_concat + sk_w
        self.bias_concat = self.bias_concat + sk_b

        self.eval_conv.weight.data = self.weight_concat.contiguous()
        self.eval_conv.bias.data = self.bias_concat.contiguous()  # type: ignore

    def forward(self, x):
        self.update_params()
        return self.eval_conv(x)


class SPAB(nn.Module):
    def __init__(self, in_channels, end: bool = False):
        super().__init__()
        self.in_channels = in_channels
        self.c1_r = Conv3XC(in_channels, in_channels, gain1=2, s=1)
        self.c2_r = Conv3XC(in_channels, in_channels, gain1=2, s=1)
        self.c3_r = Conv3XC(in_channels, in_channels, gain1=2, s=1)
        self.act1 = torch.nn.SiLU(inplace=True)
        self.end = end

    def forward(self, x):
        out1 = self.c1_r(x)
        out1_act = self.act1(out1)

        out2 = self.c2_r(out1_act)
        out2_act = self.act1(out2)

        out3 = self.c3_r(out2_act)

        sim_att = torch.sigmoid(out3) - 0.5
        out = (out3 + x) * sim_att
        if self.end:
            return out, out1
        return out


class SPAN(nn.Module):
    """
    Swift Parameter-free Attention Network for Efficient Super-Resolution
    """

    hyperparameters = {}

    def __init__(
        self,
        *,
        num_in_ch: int,
        num_out_ch: int,
        feature_channels=48,
        upscale=4,
        norm=True,
        img_range=255.0,
        rgb_mean=(0.4488, 0.4371, 0.4040),
    ):
        super().__init__()

        self.in_channels = num_in_ch
        self.out_channels = num_out_ch
        self.img_range = img_range
        self.mean = torch.Tensor(rgb_mean).view(1, 3, 1, 1)

        self.no_norm: torch.Tensor | None
        if not norm:
            self.register_buffer('no_norm', torch.zeros(1))
        else:
            self.no_norm = None

        self.conv_1 = Conv3XC(self.in_channels, feature_channels, gain1=2, s=1)
        self.block_1 = SPAB(feature_channels)
        self.block_2 = SPAB(feature_channels)
        self.block_3 = SPAB(feature_channels)
        self.block_4 = SPAB(feature_channels)
        self.block_5 = SPAB(feature_channels)
        self.block_6 = SPAB(feature_channels, end=True)

        self.conv_cat = conv_layer(feature_channels * 4, feature_channels, kernel_size=1, bias=True)
        self.conv_2 = Conv3XC(feature_channels, feature_channels, gain1=2, s=1)

        self.upsampler = pixelshuffle_block(feature_channels, self.out_channels, upscale_factor=upscale)

    @property
    def is_norm(self):
        return self.no_norm is None

    def forward(self, x):
        if self.is_norm:
            self.mean = self.mean.type_as(x)
            x = (x - self.mean) * self.img_range

        out_feature = self.conv_1(x)

        out_b1 = self.block_1(out_feature)
        out_b2 = self.block_2(out_b1)
        out_b3 = self.block_3(out_b2)

        out_b4 = self.block_4(out_b3)
        out_b5 = self.block_5(out_b4)
        out_b6, out_b5_2 = self.block_6(out_b5)

        out_b6 = self.conv_2(out_b6)
        out = self.conv_cat(torch.cat([out_feature, out_b6, out_b1, out_b5_2], 1))
        output = self.upsampler(out)

        return output
